#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

import time
from datetime import datetime, timedelta
from typing import Optional, Dict, Any

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_ASSISTANT,
    MODE_CHAT,
    MODE_EXPERT,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)

from .context import BridgeContext
from .worker import BridgeWorker

class Bridge:
    def __init__(self, window=None):
        """
        Provider bridge

        :param window: Window instance
        """
        self.window = window
        self.last_call = None  # last API call time, for throttling
        self.last_context = None  # last context
        self.last_context_quick = None  # last context for quick call
        self.sync_modes = [
            MODE_ASSISTANT,
            MODE_EXPERT,
        ]
        self.worker = None

    def request(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Make request to provider

        :param context: Bridge context
        :param extra: extra data
        """
        if self.window.controller.kernel.stopped():
            return False

        allowed_model_change = [MODE_VISION]
        is_virtual = False

        self.window.stateChanged.emit(self.window.STATE_BUSY)  # set busy

        # debug
        self.window.core.debug.info("[bridge] Request...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))

        # get data
        ctx = context.ctx
        prompt = context.prompt
        mode = context.mode
        model = context.model  # model instance, not ID
        base_mode = mode
        context.parent_mode = base_mode  # store base mode

        # get agent or expert internal sub-mode
        if base_mode in [MODE_AGENT, MODE_EXPERT]:
            is_virtual = True
            sub_mode = None  # inline switch to sub-mode, because agent is a virtual mode only
            if base_mode == MODE_AGENT:
                sub_mode = self.window.core.agents.legacy.get_mode()
            elif base_mode == MODE_EXPERT:
                sub_mode = self.window.core.experts.get_mode()
            if sub_mode is not None and sub_mode != "_":
                mode = sub_mode

        # check if model is supported by selected mode - if not, then try to use supported mode
        if model is not None:
            if not model.is_supported(mode):  # check selected mode
                mode = self.window.core.models.get_supported_mode(model, mode)  # switch

        if mode == MODE_LLAMA_INDEX and base_mode != MODE_LLAMA_INDEX:
            context.idx_mode = MODE_CHAT

        if is_virtual:
            if mode == MODE_LLAMA_INDEX:  # after switch
                idx = self.window.core.agents.legacy.get_idx()  # get index, idx is common for agent and expert
                if idx is not None and idx != "_":
                    context.idx = idx
                    self.window.core.debug.info("[agent/expert] Using index: " + idx)
                else:
                    context.idx = None  # don't use index

        # inline: internal mode switch if needed
        mode = self.window.controller.mode.switch_inline(mode, ctx, prompt)
        context.mode = mode

        # inline: model switch
        if mode in allowed_model_change:
            context.model = self.window.controller.model.switch_inline(mode, model)

        # debug
        self.window.core.debug.info("[bridge] After inline...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))

        self.apply_rate_limit()  # apply RPM limit
        self.last_context = context  # store last context for call (debug)

        if extra is None:
            extra = {}

        # async worker
        self.worker = self.get_worker()
        self.worker.context = context
        self.worker.extra = extra
        self.worker.mode = mode

        # some modes must be called synchronously
        if mode in self.sync_modes:
            self.window.core.debug.info("[bridge] Starting worker (sync)...")
            self.worker.run()
            return True

        # async call
        self.window.core.debug.info("[bridge] Starting worker (async)...")
        self.window.threadpool.start(self.worker)
        return True

    def request_next(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> bool:
        """
        Make next call to provider (loop next step)

        :param context: Bridge context
        :param extra: extra data
        """
        if self.window.controller.kernel.stopped():
            return False

        if extra is None:
            extra = {}

        # async worker
        self.worker = self.get_worker()
        self.worker.context = context
        self.worker.extra = extra
        self.worker.mode = "loop_next"

        # async call
        self.window.threadpool.start(self.worker)
        return True

    def call(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None
    ) -> str:
        """
        Make quick call to provider and get response content

        :param context: Bridge context
        :param extra: extra data
        :return: response content
        """
        if self.window.controller.kernel.stopped():
            return ""

        self.window.core.debug.info("[bridge] Call...")
        if self.window.core.debug.enabled():
            if self.window.core.config.get("log.ctx"):
                debug = {k: str(v) for k, v in context.to_dict().items()}
                self.window.core.debug.debug(str(debug))

        self.last_context_quick = context  # store last context for quick call (debug)

        if context.model is not None:
            # check if model is supported by chat API, if not then try to use llama-index or langchain call
            if not context.model.is_supported(MODE_CHAT) and not context.model.is_supported(MODE_RESEARCH):

                # tmp switch to: llama-index
                if context.model.is_supported(MODE_LLAMA_INDEX):
                    context.stream = False  # force disable stream
                    ctx = context.ctx  # output will be filled in query
                    ctx.input = context.prompt
                    try:
                        res = self.window.core.idx.chat.chat(
                            context=context,
                            extra=extra,
                        )
                        if res:
                            return ctx.output  # response text is in ctx.output
                    except Exception as e:
                        self.window.core.debug.error("Error in Llama-index quick call: " + str(e))
                        self.window.core.debug.error(e)
                    return ""

                # tmp switch to: langchain
                elif context.model.is_supported(MODE_LANGCHAIN):
                    context.stream = False
                    ctx = context.ctx
                    ctx.input = context.prompt
                    try:
                        res = self.window.core.chain.chat(
                            context=context,
                            extra=extra,
                        )
                        if res:
                            return ctx.output  # response text is in ctx.output
                    except Exception as e:
                        self.window.core.debug.error("Error in Langchain quick call: " + str(e))
                        self.window.core.debug.error(e)
                    return ""

        # if model is research model, then switch to research / Perplexity endpoint
        if context.mode is None or context.mode == MODE_CHAT:
            if context.model is not None:
                if not context.model.is_supported(MODE_CHAT):
                    if context.model.is_supported(MODE_RESEARCH):
                        context.mode = MODE_RESEARCH

        # default: OpenAI API call
        return self.window.core.gpt.quick_call(
            context=context,
            extra=extra,
        )

    def get_worker(self) -> BridgeWorker:
        """
        Prepare async worker

        :return: BridgeWorker
        """
        worker = BridgeWorker()
        worker.window = self.window
        worker.signals.response.connect(self.window.controller.kernel.listener)
        return worker

    def apply_rate_limit(self):
        """Apply API calls RPM limit"""
        max_per_minute = 60
        if self.window.core.config.has("max_requests_limit"):
            max_per_minute = int(self.window.core.config.get("max_requests_limit")) # per minute
        if max_per_minute <= 0:
            return
        interval = timedelta(minutes=1) / max_per_minute
        now = datetime.now()
        if self.last_call is not None:
            time_since_last_call = now - self.last_call
            if time_since_last_call < interval:
                sleep_time = (interval - time_since_last_call).total_seconds()
                self.window.core.debug.debug("RPM limit: sleep for {} seconds".format(sleep_time))
                time.sleep(sleep_time)
        self.last_call = now



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import json
import re
from typing import Optional, Dict, List, Any

from pygpt_net.core.bridge.context import BridgeContext

from pygpt_net.core.events import ControlEvent, AppEvent, KernelEvent


class Voice:
    def __init__(self, window=None):
        """
        Voice access core

        :param window: Window instance
        """
        self.window = window
        self.commands = {
            ControlEvent.APP_STATUS: "Get the current application status",
            ControlEvent.APP_EXIT: "Exit the application",
            ControlEvent.AUDIO_OUTPUT_ENABLE: "Enable audio output",
            ControlEvent.AUDIO_OUTPUT_DISABLE: "Disable audio output",
            ControlEvent.AUDIO_INPUT_ENABLE: "Enable audio input",
            ControlEvent.AUDIO_INPUT_DISABLE: "Disable audio input",
            ControlEvent.CALENDAR_ADD: "Add a memo to the calendar",
            ControlEvent.CALENDAR_CLEAR: "Clear memos from calendar",
            ControlEvent.CALENDAR_READ: "Read the calendar memos",
            ControlEvent.CAMERA_ENABLE: "Enable the camera",
            ControlEvent.CAMERA_DISABLE: "Disable the camera",
            ControlEvent.CAMERA_CAPTURE: "Capture image from camera",
            ControlEvent.CMD_CONFIRM: "Confirmation of the command",
            ControlEvent.CMD_LIST: "Get available commands list",
            ControlEvent.CTX_NEW: "Create a new context",
            ControlEvent.CTX_PREV: "Go to the previous context",
            ControlEvent.CTX_NEXT: "Go to the next context",
            ControlEvent.CTX_LAST: "Go to the latest context",
            ControlEvent.CTX_INPUT_FOCUS: "Focus on the input",
            ControlEvent.CTX_INPUT_SEND: "Send the input",
            ControlEvent.CTX_INPUT_CLEAR: "Clear the input",
            ControlEvent.CTX_CURRENT: "Get current conversation info",
            ControlEvent.CTX_STOP: "Stop executing current action",
            ControlEvent.CTX_ATTACHMENTS_CLEAR: "Clear the attachments",
            ControlEvent.CTX_READ_LAST: "Read the last conversation entry",
            ControlEvent.CTX_READ_ALL: "Read the whole conversation",
            ControlEvent.CTX_RENAME: "Rename current context",
            ControlEvent.CTX_SEARCH_STRING: "Search for a conversation",
            ControlEvent.CTX_SEARCH_CLEAR: "Clear the search results",
            ControlEvent.INPUT_SEND: "Send the message to input",
            ControlEvent.INPUT_APPEND: "Append message to current input without sending it",
            ControlEvent.MODE_CHAT: "Switch to Chat mode",
            ControlEvent.MODE_LLAMA_INDEX: "Switch to Chat with Files (LlamaIndex) mode",
            ControlEvent.MODE_NEXT: "Switch to the next mode",
            ControlEvent.MODE_PREV: "Switch to the previous mode",
            ControlEvent.MODEL_NEXT: "Switch to the next model",
            ControlEvent.MODEL_PREV: "Switch to the previous model",
            ControlEvent.NOTE_ADD: "Add note to notepad",
            ControlEvent.NOTEPAD_CLEAR: "Clear notepad contents",
            ControlEvent.NOTEPAD_READ: "Read current notepad contents",
            ControlEvent.PRESET_NEXT: "Switch to the next preset",
            ControlEvent.PRESET_PREV: "Switch to the previous preset",
            ControlEvent.TAB_CHAT: "Switch to the chat tab",
            ControlEvent.TAB_CALENDAR: "Switch to the calendar tab",
            ControlEvent.TAB_DRAW: "Switch to the draw (painter) tab",
            ControlEvent.TAB_FILES: "Switch to the files tab",
            ControlEvent.TAB_NOTEPAD: "Switch to the notepad tab",
            ControlEvent.TAB_NEXT: "Switch to the next tab",
            ControlEvent.TAB_PREV: "Switch to the previous tab",
            ControlEvent.VOICE_MESSAGE_START: "Start listening for voice input",
            ControlEvent.VOICE_MESSAGE_STOP: "Stop listening for voice input",
            ControlEvent.VOICE_MESSAGE_TOGGLE: "Toggle listening for voice input",
        }
        # events that should not be cached (dynamic data)
        self.cache_disabled_events = [
            ControlEvent.APP_STATUS,
            ControlEvent.CALENDAR_READ,
            ControlEvent.CTX_CURRENT,
            ControlEvent.CTX_RENAME,
            ControlEvent.CTX_READ_LAST,
            ControlEvent.CTX_READ_ALL,
            ControlEvent.CTX_SEARCH_STRING,
            ControlEvent.NOTEPAD_READ,
            AppEvent.CTX_SELECTED,
            AppEvent.MODE_SELECTED,
            AppEvent.MODEL_SELECTED,
            AppEvent.PRESET_SELECTED,
            AppEvent.TAB_SELECTED,  # notepad tabs can have custom names
        ]

    def get_commands(self) -> Dict[str, str]:
        """
        Get available commands

        :return: commands
        """
        cmds = {}
        for k, v in self.commands.items():
            if not self.is_blacklisted(k):
                cmds[k] = v
        return cmds

    def get_commands_string(self, values: bool = False) -> str:
        """
        Get available commands as string

        :param values: if True, then only values are returned
        :return: commands list as string
        """
        if values:
            numbered = {}
            for i, v in enumerate(self.get_commands().values()):
                numbered[str(i+1)] = v
            return "\n".join([f"{k}) {v}" for k, v in numbered.items()])
        else:
            return "\n".join([f"{k} = {v}" for k, v in self.get_commands().items()])

    def get_prompt(self, text: str) -> str:
        """
        Get prompt for voice command recognition

        :param text: text to be spoken
        :return: prompt
        """
        commands_str = self.get_commands_string()
        prompt = """
        Recognize the voice command and select the corresponding command from the list below. 
        Return the chosen command ID as a JSON string in the following syntax:

        {"cmd": "command_id", "params": "optional message"}
        
        If no command matches the voice request, then return "unknown" as the command ID.
        If user provide additional message, it should be extracted from voice command and included (WITHOUT the command part) in the JSON response in "params" param.
        
        Important: Only the JSON specified above should be returned in the response, without any additional text.
        
        Available command IDs (with descriptions):
        ----------------------
        """+commands_str+"""        
        
        User's voice input to recognize:
        
        """+text

        return prompt.strip()

    def get_inline_prompt(self, prefix: Optional[str] = None) -> str:
        """
        Get prompt for inline voice command recognition

        :return: prompt
        """
        prefix_addon = ""
        if prefix is not None and prefix.strip() != "":
            prefix_addon = ' or action prefixed via "{}" (or similar)'.format(prefix)

        prompt = """
           If user provides the voice action (as a text)"""+prefix_addon+""" then recognize this voice action and execute the corresponding voice action from the enum list.
           If no defined actions matches the requested action, then return "unknown" as the action. Do not try to execute any other voice commands.
           If user provide additional message, it should be extracted from query and included (WITHOUT the action part) in the "args" param.
           IMPORTANT: Only execute actions from provided list and remember than list describes only voice actions, not real commands, so don't execute any commands from this list separately.
           When calling voice action, the action should be returned in a JSON string, without any additional text.
           JSON schema required for voice action is below:
           {"cmd": "voice_cmd", "params": {"action": "action_id", "args": "optional message"}}
           """

        return prompt.strip()

    def extract_json(self, text: str) -> List[Dict[str, Any]]:
        """
        Extract JSON from text

        :param text: text
        :return: JSON
        """
        json_pattern = r'(\{.*?\})'
        cmds = []
        enabled_cmds = self.get_commands()
        matches = re.findall(json_pattern, text, re.DOTALL)
        if len(matches) > 0:
            for match in matches:
                try:
                    data = json.loads(match)
                    if "cmd" in data:
                        cmd = data["cmd"]
                        if cmd in enabled_cmds:
                            item = {
                                "cmd": cmd,
                                "params": data.get("params", "")
                            }
                            cmds.append(item)
                        else:
                            cmds.append({
                                "cmd": "unrecognized",
                                "params": ""
                            })
                except Exception as e:
                    self.window.core.debug.log(e)
        return cmds

    def recognize_commands(self, text: str) -> List[Dict[str, Any]]:
        """
        Recognize voice command

        :param text: voice input
        :return: recognized command
        """
        prompt = self.get_prompt(text)
        model = self.window.core.models.from_defaults()
        tmp_model = self.window.core.config.get("access.voice_control.model", "gpt-3.5-turbo")
        if self.window.core.models.has(tmp_model):
            model = self.window.core.models.get(tmp_model)
        bridge_context = BridgeContext(
            prompt=prompt,
            system_prompt="You are a helpful assistant",
            model=model,  # model instance
            max_tokens=0,
            temperature=0.0,
        )
        event = KernelEvent(KernelEvent.CALL, {
            'context': bridge_context,
            'extra': {},
        })
        self.window.dispatch(event)
        response = event.data.get('response')
        if response is None or response == "":
            return []

        return self.extract_json(response)

    def is_muted(self, action: str) -> bool:
        """
        Check if audio event is blacklisted

        :param action: event name
        :return: True if blacklisted
        """
        data = self.window.core.config.get("access.audio.event.speech.disabled")
        if data is None or not isinstance(data, list):
            return False
        for item in data:
            if "muted_action" in item and item["muted_action"] == action:
                return True
        return False

    def is_blacklisted(self, action: str) -> bool:
        """
        Check if audio control action is blacklisted

        :param action: action name
        :return: True if blacklisted
        """
        data = self.window.core.config.get("access.voice_control.blacklist")
        if data is None or not isinstance(data, list):
            return False
        for item in data:
            if "disabled_action" in item and item["disabled_action"] == action:
                return True
        return False

    def cache_disabled(self, action: str) -> bool:
        """
        Check if audio control action should be cached

        :param action: action name
        :return: True if cached
        """
        return action in self.cache_disabled_events



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 03:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QVBoxLayout, QWidget

from pygpt_net.ui.widget.element.labels import TitleLabel
from pygpt_net.ui.widget.lists.model_combo import ModelCombo
from pygpt_net.utils import trans


class Model:
    def __init__(self, window=None):
        """
        Toolbox UI

        :param window: Window instance
        """
        self.window = window
        self.id = 'prompt.model'

    def setup(self) -> QWidget:
        """
        Setup models

        :return: QWidget7
        """
        widget = QWidget()
        widget.setLayout(self.setup_list())
        return widget

    def setup_list(self) -> QVBoxLayout:
        """
        Setup models list

        :return: QVBoxLayout
        """
        label_key = self.id + '.label'

        self.window.ui.nodes[label_key] = TitleLabel(trans("toolbox.model.label"))
        self.window.ui.nodes[self.id] = ModelCombo(self.window, self.id)

        header_layout = QVBoxLayout()
        header_layout.addWidget(self.window.ui.nodes[label_key])
        header_layout.setContentsMargins(5, 0, 0, 0)

        layout = QVBoxLayout()
        layout.addLayout(header_layout)
        layout.addWidget(self.window.ui.nodes[self.id])
        layout.addStretch()
        layout.setContentsMargins(2, 5, 5, 5)

        return layout



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.24 22:00:00                  #
# ================================================== #

from PySide6.QtCore import (
    Qt, QSize, QPoint, QPointF, QRectF,
    QEasingCurve, QPropertyAnimation, QSequentialAnimationGroup,
    Slot, Property, QEvent)

from PySide6.QtWidgets import QCheckBox
from PySide6.QtGui import QColor, QBrush, QPaintEvent, QPen, QPainter, QPalette


class AnimToggle(QCheckBox):

    # Based on widget from: https://www.pythonguis.com/tutorials/pyside6-animated-widgets/

    _transparent_pen = QPen(Qt.transparent)
    _light_grey_pen = QPen(Qt.lightGray)

    def __init__(self, title="", parent=None):
        super().__init__()
        self.setText(title)
        self.option = None

        # Initialize the colors based on the current palette
        self.updateColors()

        # Setup the rest of the widget.
        self.setContentsMargins(8, 0, 8, 0)
        self.setMaximumWidth(58)
        self._handle_position = 0
        self._pulse_radius = 0

        # Setup animations
        self.animation = QPropertyAnimation(self, b"handle_position", self)
        self.animation.setEasingCurve(QEasingCurve.InOutCubic)
        self.animation.setDuration(200)  # time in ms

        self.pulse_anim = QPropertyAnimation(self, b"pulse_radius", self)
        self.pulse_anim.setDuration(350)  # time in ms
        self.pulse_anim.setStartValue(10)
        self.pulse_anim.setEndValue(20)

        self.animations_group = QSequentialAnimationGroup()
        self.animations_group.addAnimation(self.animation)
        self.animations_group.addAnimation(self.pulse_anim)

        self.stateChanged.connect(self.setup_animation)

    def updateColors(self):
        # Get colors from the palette.
        palette = self.palette()
        #bar_color = QPen(Qt.lightGray).color()
        bar_color = QColor("#969696")
        #bar_checked_color = palette.color(QPalette.Dark)
        bar_checked_color = QColor("#969696")
        bar_color.setAlpha(0x80)
        handle_color = palette.color(QPalette.Dark)
        handle_checked_color = palette.color(QPalette.ButtonText)

        # Create semi-transparent colors for the pulse effect
        pulse_unchecked_color = bar_color.darker(110)
        pulse_checked_color = bar_checked_color.lighter(130)
        pulse_unchecked_color.setAlpha(0x44)
        pulse_checked_color.setAlpha(0x44)

        # Now set the brushes
        self._bar_brush = QBrush(bar_color)
        self._bar_checked_brush = QBrush(bar_checked_color)
        self._handle_brush = QBrush(handle_color)
        self._handle_checked_brush = QBrush(handle_checked_color)
        self._pulse_unchecked_animation = QBrush(pulse_unchecked_color)
        self._pulse_checked_animation = QBrush(pulse_checked_color)

        # Update the widget to redraw with new colors
        self.update()

    def event(self, e):
        if e.type() == QEvent.PaletteChange:
            self.updateColors()
        return super().event(e)

    def sizeHint(self):
        return QSize(58, 45)

    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    @Slot(int)
    def setup_animation(self, value):
        self.animations_group.stop()
        if value:
            self.animation.setEndValue(1)
        else:
            self.animation.setEndValue(0)
        self.animations_group.start()

    def paintEvent(self, e: QPaintEvent):
        contRect = self.contentsRect()
        handleRadius = round(0.24 * contRect.height())

        p = QPainter(self)
        p.setRenderHint(QPainter.Antialiasing)
        p.setPen(self._transparent_pen)

        barRect = QRectF(
            0, 0,
            contRect.width() - handleRadius, 0.40 * contRect.height()
        )
        barRect.moveCenter(contRect.center())
        rounding = barRect.height() / 2

        # The handle will move along this line
        trailLength = contRect.width() - 2 * handleRadius

        xPos = contRect.x() + handleRadius + trailLength * self._handle_position

        if self.pulse_anim.state() == QPropertyAnimation.Running:
            p.setBrush(
                self._pulse_checked_animation if
                self.isChecked() else self._pulse_unchecked_animation)
            p.drawEllipse(QPointF(xPos, barRect.center().y()),
                          self._pulse_radius, self._pulse_radius)

        if self.isChecked():
            p.setBrush(self._bar_checked_brush)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setBrush(self._handle_checked_brush)
        else:
            p.setBrush(self._bar_brush)
            p.drawRoundedRect(barRect, rounding, rounding)
            p.setPen(self._light_grey_pen)
            p.setBrush(self._handle_brush)

        p.drawEllipse(
            QPointF(xPos, barRect.center().y()),
            handleRadius, handleRadius)

        p.end()

    @Property(float)
    def handle_position(self):
        return self._handle_position

    @handle_position.setter
    def handle_position(self, pos):
        # Update the position and redraw
        self._handle_position = pos
        self.update()

    @Property(float)
    def pulse_radius(self):
        return self._pulse_radius

    @pulse_radius.setter
    def pulse_radius(self, pos):
        # Update the pulse radius and redraw
        self._pulse_radius = pos
        self.update()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.23 00:00:00                  #
# ================================================== #

from PySide6.QtGui import QAction, QIcon, QResizeEvent, Qt
from PySide6.QtWidgets import QMenu

from pygpt_net.ui.widget.lists.base import BaseList
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class AttachmentCtxList(BaseList):
    def __init__(self, window=None, id=None):
        """
        Ctx Attachments menu

        :param window: main window
        :param id: input id
        """
        super(AttachmentCtxList, self).__init__(window)
        self.window = window
        self.id = id
        self.doubleClicked.connect(self.dblclick)
        self.setHeaderHidden(False)
        self.clicked.disconnect(self.click)

        self.header = self.header()
        self.header.setStretchLastSection(False)

        self.column_proportion = 0.3
        self.adjustColumnWidths()

    def adjustColumnWidths(self):
        total_width = self.width()
        first_column_width = int(total_width * self.column_proportion)
        self.setColumnWidth(0, first_column_width)
        for column in range(1, 4):
            self.setColumnWidth(column, (total_width - first_column_width) // (4 - 1))

    def resizeEvent(self, event: QResizeEvent):
        super().resizeEvent(event)
        self.adjustColumnWidths()

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: mouse event
        """
        if event.buttons() == Qt.LeftButton:
            index = self.indexAt(event.pos())
            if index.isValid():
                self.window.controller.assistant.files.select(index.row())
        super(AttachmentCtxList, self).mousePressEvent(event)

    def click(self, val):
        """
        Click event

        :param val: click event
        """
        pass

    def dblclick(self, val):
        """
        Double click event

        :param val: double click event
        """
        self.window.controller.assistant.files.select(val.row())

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: context menu event
        """
        actions = {}

        item = self.indexAt(event.pos())
        idx = item.row()

        has_file = False
        has_src = False
        has_dest = False

        if idx >= 0:
            has_file = self.window.controller.chat.attachment.has_file_by_idx(idx)
            has_src = self.window.controller.chat.attachment.has_src_by_idx(idx)
            has_dest = self.window.controller.chat.attachment.has_dest_by_idx(idx)

        actions['open'] = QAction(QIcon(":/icons/view.svg"), trans('action.open'), self)
        actions['open'].triggered.connect(
            lambda: self.action_open(event)
        )
        actions['open_dir_src'] = QAction(QIcon(":/icons/folder.svg"), trans('action.open_dir_src'), self)
        actions['open_dir_src'].triggered.connect(
            lambda: self.action_open_dir_src(event)
        )
        actions['open_dir_dest'] = QAction(QIcon(":/icons/folder.svg"), trans('action.open_dir_storage'), self)
        actions['open_dir_dest'].triggered.connect(
            lambda: self.action_open_dir_dest(event)
        )

        actions['delete'] = QAction(QIcon(":/icons/delete.svg"), trans('action.delete'), self)
        actions['delete'].triggered.connect(
            lambda: self.action_delete(event)
        )

        menu = QMenu(self)
        if has_file:
            menu.addAction(actions['open'])
        if has_src:
            menu.addAction(actions['open_dir_src'])
        if has_dest:
            menu.addAction(actions['open_dir_dest'])
        menu.addAction(actions['delete'])

        if idx >= 0:
            self.window.controller.chat.attachment.select(item.row())
            menu.exec_(event.globalPos())

    def action_delete(self, event):
        """
        Delete action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.chat.attachment.delete_by_idx(idx)

    def action_open(self, event):
        """
        Open action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.chat.attachment.open_by_idx(idx)

    def action_open_dir_src(self, event):
        """
        Open source directory action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.chat.attachment.open_dir_src_by_idx(idx)

    def action_open_dir_dest(self, event):
        """
        Open destination directory action handler

        :param event: mouse event
        """
        item = self.indexAt(event.pos())
        idx = item.row()
        if idx >= 0:
            self.window.controller.chat.attachment.open_dir_dest_by_idx(idx)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.17 03:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QHBoxLayout, QWidget, QComboBox

class BaseListCombo(QWidget):
    def __init__(self, window=None, id: str = None):
        """
        Base list combo

        :param window: main window
        :param id: option id
        """
        super(BaseListCombo, self).__init__(window)
        self.window = window
        self.id = id
        self.value = None
        self.keys = []
        self.title = ""
        self.real_time = False
        self.combo = QComboBox()
        self.combo.currentIndexChanged.connect(self.on_combo_change)
        self.current_id = None
        self.initialized = False
        self.locked = False

        # add items
        self.update()

        self.layout = QHBoxLayout()
        self.layout.addWidget(self.combo)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)
        self.fit_to_content()
        self.initialized = True

    def update(self):
        """Prepare items"""
        if type(self.keys) is list:
            for item in self.keys:
                if type(item) is dict:
                    for key, value in item.items():
                        self.combo.addItem(value, key)
                else:
                    self.combo.addItem(item, item)
        elif type(self.keys) is dict:
            for key, value in self.keys.items():
                self.combo.addItem(value, key)

    def set_value(self, value):
        """
        Set value

        :param value: value
        """
        self.locked = True
        index = self.combo.findData(value)
        if index != -1:
            self.combo.setCurrentIndex(index)
        self.locked = False

    def get_value(self):
        """
        Get value

        :return: value
        """
        return self.current_id

    def has_key(self, name: str) -> bool:
        """
        Check if combo has key

        :param name: key name
        :return:
        """
        if isinstance(self.keys, list):
            for key in self.keys:
                if isinstance(key, dict):
                    if name in key:
                        return True
                elif name == key:
                    return True
        elif isinstance(self.keys, dict):
            if name in self.keys:
                return True
        return False

    def set_keys(self, keys):
        """
        Set keys

        :param keys: keys
        """
        self.locked = True
        self.keys = keys
        self.combo.clear()
        self.update()
        self.locked = False

    def on_combo_change(self, index):
        """
        On combo change

        :param index: combo index
        """
        if not self.initialized or self.locked:
            return
        self.current_id = self.combo.itemData(index)

    def fit_to_content(self):
        """Fit to content"""
        self.combo.setSizeAdjustPolicy(QComboBox.AdjustToContents)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "serial_port",
            type="text",
            value="/dev/ttyUSB0",
            label="USB port",
            description="USB port name, e.g. /dev/ttyUSB0, /dev/ttyACM0, COM3",
            min=1,
            max=None,
        )
        plugin.add_option(
            "serial_bps",
            type="int",
            value=9600,
            label="Connection speed (baudrate, bps)",
            description="Port connection speed, in bps, default: 9600",
            min=1,
            max=None,
        )
        plugin.add_option(
            "timeout",
            type="int",
            value=1,
            label="Timeout",
            description="Timeout in seconds, default: 1",
            min=0,
            max=None,
        )
        plugin.add_option(
            "sleep",
            type="int",
            value=2,
            label="Sleep",
            description="Sleep in seconds after connection, default: 2",
            min=0,
            max=None,
        )

        # commands
        plugin.add_cmd(
            "serial_send",
            instruction="send text command to USB port",
            params=[
                {
                    "name": "command",
                    "type": "str",
                    "description": "command",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Send text commands to USB port",
        )
        plugin.add_cmd(
            "serial_send_bytes",
            instruction="send raw bytes to USB port",
            params=[
                {
                    "name": "bytes",
                    "type": "int",
                    "description": "bytes",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Send raw bytes to USB port",
        )
        plugin.add_cmd(
            "serial_read",
            instruction="read data from serial port in seconds duration",
            params=[
                {
                    "name": "duration",
                    "type": "int",
                    "description": "duration",
                    "required": True,
                },
            ],
            enabled=True,
            description="Enable: Read data from USB port",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

import os

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)

class PresetsDebug:
    def __init__(self, window=None):
        """
        Presets debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'presets'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)

        self.window.core.debug.add(
            self.id, 'Options',
            str(self.window.controller.presets.editor.get_options())
        )

        # presets
        for key in list(dict(self.window.core.presets.items)):
            preset = self.window.core.presets.items[key]
            path = os.path.join(self.window.core.config.path, 'presets', key + '.json')
            data = {
                'id': key,
                'file': path,
                'name': preset.name,
                'ai_name': preset.ai_name,
                'user_name': preset.user_name,
                'prompt': preset.prompt,
                MODE_CHAT: preset.chat,
                MODE_COMPLETION: preset.completion,
                MODE_IMAGE: preset.img,
                MODE_VISION: preset.vision,
                MODE_LANGCHAIN: preset.langchain,
                MODE_ASSISTANT: preset.assistant,
                MODE_LLAMA_INDEX: preset.llama_index,
                MODE_AGENT: preset.agent,
                MODE_AGENT_LLAMA: preset.agent_llama,
                MODE_EXPERT: preset.expert,
                MODE_AUDIO: preset.audio,
                MODE_RESEARCH: preset.research,
                'temperature': preset.temperature,
                'version': preset.version,
            }
            self.window.core.debug.add(self.id, str(key), str(data))

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 17:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QWidget
from PySide6.QtCore import Qt, QPropertyAnimation, QRectF, Property, QSize, QEasingCurve, QPauseAnimation, \
    QSequentialAnimationGroup
from PySide6.QtGui import QPainter, QColor, QBrush

class Loader(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._size = 10
        self._scale = 1.0
        self._base_size = 30
        self._color = QColor(100, 100, 100)
        self.setFixedSize(self._base_size, self._base_size)
        self.animation = QPropertyAnimation(self, b"scale")
        self.animation.setDuration(1000)
        self.animation.setStartValue(0.8)
        self.animation.setEndValue(1.2)
        self.animation.setLoopCount(-1)
        self.started = False
        # self.animation.start()

    def start_anim(self):
        """Start animation."""
        if self.started:
            return
        self.animation.start()
        self.started = True

    def stop_anim(self):
        """Stop animation."""
        self.started = False
        self.animation.stop()

    def paintEvent(self, event):
        """
        Paint event

        :param event: event
        """
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)

        brush = QBrush(self._color)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        scaled_size = self._size * self._scale
        x = (self.width() - scaled_size) / 2
        y = (self.height() - scaled_size) / 2

        painter.drawEllipse(QRectF(x, y, scaled_size, scaled_size))

    def sizeHint(self):
        """Size hint"""
        return QSize(self._base_size, self._base_size)

    def getScale(self):
        """
        Get scale

        :return: scale
        """
        return self._scale

    def setScale(self, scale):
        """
        Set scale

        :param scale: scale
        """
        self._scale = scale
        self.update()

    scale = Property(float, getScale, setScale)


class Loading(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.square_size = 5
        self.total_width = 100
        self.spacing = (self.total_width - 3 * self.square_size) / 2
        self._pos1 = self._pos2 = self._pos3 = 0
        self.started = False
        self.initialized = False
        # self.start_anim()

    def sizeHint(self):
        """
        Size hint

        :return: size
        """
        return QSize(self.total_width, 5)

    def start_anim(self):
        """Start animation."""
        if self.started:
            return
        self.anim1 = self.create_animation(b"pos1", 0)
        self.anim2 = self.create_animation(b"pos2", 200)
        self.anim3 = self.create_animation(b"pos3", 400)
        self.anim1.start()
        self.anim2.start()
        self.anim3.start()
        self.started = True
        self.initialized = True

    def stop_anim(self):
        """Stop animation."""
        self.started = False
        if self.initialized:
            self.anim1.stop()
            self.anim2.stop()
            self.anim3.stop()

    def create_animation(self, property_name, delay):
        """
        Create animation

        :param property_name: property name
        :param delay: delay
        """
        sequence = QSequentialAnimationGroup()
        pause = QPauseAnimation(delay)
        sequence.addAnimation(pause)
        anim = QPropertyAnimation(self, property_name)
        anim.setDuration(800)
        anim.setStartValue(0)
        anim.setKeyValueAt(0.5, self.total_width - self.square_size)
        anim.setEndValue(0)
        anim.setEasingCurve(QEasingCurve.InOutQuad)
        anim.setLoopCount(-1)
        sequence.addAnimation(anim)
        sequence.setLoopCount(-1)
        return sequence

    @Property(float)
    def pos1(self):
        """
        Position 1

        :return: position
        """
        return self._pos1

    @pos1.setter
    def pos1(self, value):
        """
        Set position 1

        :param value: value
        """
        self._pos1 = value
        self.update()

    @Property(float)
    def pos2(self):
        """
        Position 2

        :return: position
        """
        return self._pos2

    @pos2.setter
    def pos2(self, value):
        """
        Set position 2

        :param value: value
        """
        self._pos2 = value
        self.update()

    @Property(float)
    def pos3(self):
        """
        Position 3

        :return: position
        """
        return self._pos3

    @pos3.setter
    def pos3(self, value):
        """
        Set position 3

        :param value: value
        """
        self._pos3 = value
        self.update()

    def paintEvent(self, event):
        """
        Paint event

        :param event: event
        """
        painter = QPainter(self)
        y = (self.height() - self.square_size) / 2

        rect1 = QRectF(self.pos1, y, self.square_size, self.square_size)
        rect2 = QRectF(self.pos2, y, self.square_size, self.square_size)
        rect3 = QRectF(self.pos3, y, self.square_size, self.square_size)

        # colors
        #painter.fillRect(rect1, QColor("#3498db"))
        #painter.fillRect(rect2, QColor("#e74c3c"))
        #painter.fillRect(rect3, QColor("#f1c40f"))

        # gray colors
        painter.fillRect(rect1, QColor("#e0e0e0"))
        painter.fillRect(rect2, QColor("#bcbcbc"))
        painter.fillRect(rect3, QColor("#959595"))

        painter.end()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

MODE_AUDIO = "audio"
MODE_CHAT = "chat"
MODE_COMPLETION = "completion"
MODE_IMAGE = "img"
MODE_VISION = "vision"
MODE_ASSISTANT = "assistant"
MODE_LANGCHAIN = "langchain"
MODE_LLAMA_INDEX = "llama_index"
MODE_AGENT = "agent"
MODE_AGENT_LLAMA = "agent_llama"
MODE_EXPERT = "expert"
MODE_RESEARCH = "research"


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.31 19:00:00                  #
# ================================================== #

import os

# disable warnings
os.environ["TRANSFORMERS_NO_ADVISORY_WARNINGS"] = "1"

from pygpt_net.launcher import Launcher

# plugins
from pygpt_net.plugin.voice_control import Plugin as VoiceControlPlugin
from pygpt_net.plugin.audio_output import Plugin as AudioOutputPlugin
from pygpt_net.plugin.audio_input import Plugin as AudioInputPlugin
from pygpt_net.plugin.cmd_api import Plugin as CmdApiPlugin
from pygpt_net.plugin.cmd_code_interpreter import Plugin as CmdCodeInterpreterPlugin
from pygpt_net.plugin.cmd_custom import Plugin as CmdCustomCommandPlugin
from pygpt_net.plugin.cmd_files import Plugin as CmdFilesPlugin
from pygpt_net.plugin.cmd_history import Plugin as CtxHistoryPlugin
from pygpt_net.plugin.cmd_mouse_control import Plugin as CmdMousePlugin
from pygpt_net.plugin.cmd_serial import Plugin as CmdSerialPlugin
from pygpt_net.plugin.cmd_system import Plugin as CmdSystemPlugin
from pygpt_net.plugin.cmd_web import Plugin as CmdWebPlugin
from pygpt_net.plugin.crontab import Plugin as CrontabPlugin
from pygpt_net.plugin.extra_prompt import Plugin as ExtraPromptPlugin
from pygpt_net.plugin.experts import Plugin as ExpertsPlugin
from pygpt_net.plugin.idx_llama_index import Plugin as IdxLlamaIndexPlugin
from pygpt_net.plugin.openai_dalle import Plugin as OpenAIDallePlugin
from pygpt_net.plugin.openai_vision import Plugin as OpenAIVisionPlugin
from pygpt_net.plugin.real_time import Plugin as RealTimePlugin
from pygpt_net.plugin.agent import Plugin as AgentPlugin
from pygpt_net.plugin.mailer import Plugin as MailerPlugin

# agents (Llama-index)
from pygpt_net.provider.agents.openai import OpenAIAgent
from pygpt_net.provider.agents.openai_assistant import OpenAIAssistantAgent
from pygpt_net.provider.agents.planner import PlannerAgent
from pygpt_net.provider.agents.react import ReactAgent

# LLM wrapper providers (langchain, llama-index, embeddings)
from pygpt_net.provider.llms.anthropic import AnthropicLLM
from pygpt_net.provider.llms.azure_openai import AzureOpenAILLM
from pygpt_net.provider.llms.deepseek_api import DeepseekApiLLM
from pygpt_net.provider.llms.google import GoogleLLM
from pygpt_net.provider.llms.hugging_face import HuggingFaceLLM
from pygpt_net.provider.llms.hugging_face_api import HuggingFaceApiLLM
from pygpt_net.provider.llms.local import LocalLLM
from pygpt_net.provider.llms.ollama import OllamaLLM
from pygpt_net.provider.llms.openai import OpenAILLM

# vector store providers (llama-index)
from pygpt_net.provider.vector_stores.chroma import ChromaProvider
from pygpt_net.provider.vector_stores.elasticsearch import ElasticsearchProvider
from pygpt_net.provider.vector_stores.pinecode import PinecodeProvider
from pygpt_net.provider.vector_stores.redis import RedisProvider
from pygpt_net.provider.vector_stores.simple import SimpleProvider

# data loader providers (llama-index)
from pygpt_net.provider.loaders.file_csv import Loader as CsvLoader
from pygpt_net.provider.loaders.file_docx import Loader as DocxLoader
from pygpt_net.provider.loaders.file_epub import Loader as EpubLoader
from pygpt_net.provider.loaders.file_excel import Loader as ExcelLoader
from pygpt_net.provider.loaders.file_html import Loader as HtmlLoader
from pygpt_net.provider.loaders.file_image_vision import Loader as ImageVisionLoader
from pygpt_net.provider.loaders.file_ipynb import Loader as IPYNBLoader
from pygpt_net.provider.loaders.file_json import Loader as JsonLoader
from pygpt_net.provider.loaders.file_markdown import Loader as MarkdownLoader
from pygpt_net.provider.loaders.file_pdf import Loader as PdfLoader
from pygpt_net.provider.loaders.file_video_audio import Loader as VideoAudioLoader
from pygpt_net.provider.loaders.file_xml import Loader as XmlLoader
from pygpt_net.provider.loaders.web_bitbucket import Loader as BitbucketRepoLoader
from pygpt_net.provider.loaders.web_chatgpt_retrieval import Loader as ChatGptRetrievalLoader
from pygpt_net.provider.loaders.web_database import Loader as DatabaseLoader
from pygpt_net.provider.loaders.web_github_issues import Loader as GithubIssuesLoader
from pygpt_net.provider.loaders.web_github_repo import Loader as GithubRepoLoader
from pygpt_net.provider.loaders.web_google_calendar import Loader as GoogleCalendarLoader
from pygpt_net.provider.loaders.web_google_docs import Loader as GoogleDocsLoader
from pygpt_net.provider.loaders.web_google_drive import Loader as GoogleDriveLoader
from pygpt_net.provider.loaders.web_google_gmail import Loader as GoogleGmailLoader
from pygpt_net.provider.loaders.web_google_keep import Loader as GoogleKeepLoader
from pygpt_net.provider.loaders.web_google_sheets import Loader as GoogleSheetsLoader
from pygpt_net.provider.loaders.web_microsoft_onedrive import Loader as MicrosoftOneDriveLoader
from pygpt_net.provider.loaders.web_rss import Loader as RssLoader
from pygpt_net.provider.loaders.web_sitemap import Loader as SitemapLoader
from pygpt_net.provider.loaders.web_twitter import Loader as TwitterLoader
from pygpt_net.provider.loaders.web_page import Loader as WebPageLoader
from pygpt_net.provider.loaders.web_yt import Loader as YouTubeLoader

# audio providers (input, output)
from pygpt_net.provider.audio_input.openai_whisper import OpenAIWhisper
from pygpt_net.provider.audio_input.openai_whisper_local import OpenAIWhisperLocal
from pygpt_net.provider.audio_input.google_speech_recognition import GoogleSpeechRecognition
from pygpt_net.provider.audio_input.google_cloud_speech_recognition import GoogleCloudSpeechRecognition
from pygpt_net.provider.audio_input.bing_speech_recognition import BingSpeechRecognition
from pygpt_net.provider.audio_output.openai_tts import OpenAITextToSpeech
from pygpt_net.provider.audio_output.ms_azure_tts import MSAzureTextToSpeech
from pygpt_net.provider.audio_output.google_tts import GoogleTextToSpeech
from pygpt_net.provider.audio_output.eleven_labs import ElevenLabsTextToSpeech

# web search engine providers
from pygpt_net.provider.web.google_custom_search import GoogleCustomSearch
from pygpt_net.provider.web.microsoft_bing import MicrosoftBingSearch

# tools
from pygpt_net.tools.indexer import IndexerTool
from pygpt_net.tools.audio_transcriber import AudioTranscriber as AudioTranscriberTool
from pygpt_net.tools.code_interpreter import CodeInterpreter as CodeInterpreterTool
from pygpt_net.tools.image_viewer import ImageViewer as ImageViewerTool
from pygpt_net.tools.media_player import MediaPlayer as MediaPlayerTool
from pygpt_net.tools.text_editor import TextEditor as TextEditorTool
from pygpt_net.tools.html_canvas import HtmlCanvas as HtmlCanvasTool

def run(**kwargs):
    """
    PyGPT launcher.

    :param kwargs: keyword arguments for launcher

    PyGPT can be extended with:

    - Custom plugins
    - Custom LLM wrappers
    - Custom vector store providers
    - Custom data loaders
    - Custom audio input providers
    - Custom audio output providers
    - Custom web search engine providers
    - Custom tools

    - You can pass custom plugin instances, LLM wrappers, vector store providers and more to the launcher.
    - This is useful if you want to extend PyGPT with your own plugins, vector storage, LLMs, or other data providers.

    First, create a custom launcher file, for example, "custom_launcher.py," and register your extensions in it.

    To register a custom plugin - create the custom launcher, e.g. "custom_launcher.py" and:

    - Pass a list with the plugin instances as the 'plugins' keyword argument.

    To register a custom LLM wrapper:

    - Pass a list with the LLM wrapper instances as the 'llms' keyword argument.

    To register a custom vector store provider:

    - Pass a list with the vector store provider instances as the 'vector_stores' keyword argument.

    To register a custom data loader:

    - Pass a list with the data loader instances as the 'loaders' keyword argument.

    To register a custom audio input provider:

    - Pass a list with the audio input provider instances as the 'audio_input' keyword argument.

    To register a custom audio output provider:

    - Pass a list with the audio output provider instances as the 'audio_output' keyword argument.

    To register a custom web provider:

    - Pass a list with the web provider instances as the 'web' keyword argument.

    To register a custom agent:

    - Pass a list with the agent instances as the 'agents' keyword argument.

    To register a custom tool:

    - Pass a list with the tool instances as the 'tools' keyword argument.

    Example:
    --------
    ::

        # custom_launcher.py

        from pygpt_net.app import run
        from plugins import CustomPlugin, OtherCustomPlugin
        from llms import CustomLLM
        from vector_stores import CustomVectorStore
        from loaders import CustomLoader
        from audio_input import CustomAudioInput
        from audio_output import CustomAudioOutput
        from web import CustomWebSearch
        from tools import CustomTool

        plugins = [
            CustomPlugin(),
            OtherCustomPlugin(),
        ]
        llms = [
            CustomLLM(),
        ]
        vector_stores = [
            CustomVectorStore(),
        ]
        loaders = [
            CustomLoader(),
        ]
        audio_input = [
            CustomAudioInput(),
        ]
        audio_output = [
            CustomAudioOutput(),
        ]
        web = [
            CustomWebSearch(),
        ]
        agents = [
            CustomAgent(),
        ]
        tools = [
            CustomTool(),
        ]

        run(
            plugins=plugins,
            llms=llms,
            vector_stores=vector_stores,
            loaders=loaders,
            audio_input=audio_input,
            audio_output=audio_output,
            web=web,
            agents=agents,
            tools=tools,
        )

    """
    # initialize app launcher
    launcher = Launcher()
    launcher.init()

    # register audio providers
    launcher.add_audio_input(OpenAIWhisper())
    launcher.add_audio_input(OpenAIWhisperLocal())
    launcher.add_audio_input(GoogleSpeechRecognition())
    launcher.add_audio_input(GoogleCloudSpeechRecognition())
    launcher.add_audio_input(BingSpeechRecognition())
    launcher.add_audio_output(OpenAITextToSpeech())
    launcher.add_audio_output(MSAzureTextToSpeech())
    launcher.add_audio_output(GoogleTextToSpeech())
    launcher.add_audio_output(ElevenLabsTextToSpeech())

    # register custom audio providers
    providers = kwargs.get('audio_input', None)
    if isinstance(providers, list):
        for provider in providers:
            launcher.add_audio_input(provider)

    providers = kwargs.get('audio_output', None)
    if isinstance(providers, list):
        for provider in providers:
            launcher.add_audio_output(provider)

    # register web providers
    launcher.add_web(GoogleCustomSearch())
    launcher.add_web(MicrosoftBingSearch())

    # register custom web providers
    providers = kwargs.get('web', None)
    if isinstance(providers, list):
        for provider in providers:
            launcher.add_web(provider)

    # register base data loaders (llama-index)
    launcher.add_loader(CsvLoader())
    launcher.add_loader(DocxLoader())
    launcher.add_loader(EpubLoader())
    launcher.add_loader(ExcelLoader())
    launcher.add_loader(HtmlLoader())
    launcher.add_loader(ImageVisionLoader())
    launcher.add_loader(IPYNBLoader())
    launcher.add_loader(JsonLoader())
    launcher.add_loader(MarkdownLoader())
    launcher.add_loader(PdfLoader())
    launcher.add_loader(VideoAudioLoader())
    launcher.add_loader(XmlLoader())
    launcher.add_loader(RssLoader())
    launcher.add_loader(SitemapLoader())
    launcher.add_loader(BitbucketRepoLoader())
    launcher.add_loader(ChatGptRetrievalLoader())
    launcher.add_loader(DatabaseLoader())
    launcher.add_loader(GithubIssuesLoader())
    launcher.add_loader(GithubRepoLoader())
    launcher.add_loader(GoogleCalendarLoader())
    launcher.add_loader(GoogleDocsLoader())
    launcher.add_loader(GoogleDriveLoader())
    launcher.add_loader(GoogleGmailLoader())
    launcher.add_loader(GoogleKeepLoader())
    launcher.add_loader(GoogleSheetsLoader())
    launcher.add_loader(MicrosoftOneDriveLoader())
    launcher.add_loader(TwitterLoader())
    launcher.add_loader(WebPageLoader())
    launcher.add_loader(YouTubeLoader())

    # register custom data loaders (llama-index)
    loaders = kwargs.get('loaders', None)
    if isinstance(loaders, list):
        for loader in loaders:
            launcher.add_loader(loader)

    # register base plugins
    launcher.add_plugin(VoiceControlPlugin())
    launcher.add_plugin(AgentPlugin())
    launcher.add_plugin(RealTimePlugin())
    launcher.add_plugin(ExpertsPlugin())
    launcher.add_plugin(ExtraPromptPlugin())
    launcher.add_plugin(AudioInputPlugin())
    launcher.add_plugin(AudioOutputPlugin())
    launcher.add_plugin(CmdWebPlugin())
    launcher.add_plugin(CmdFilesPlugin())
    launcher.add_plugin(CmdCodeInterpreterPlugin())
    launcher.add_plugin(CmdSystemPlugin())
    launcher.add_plugin(CmdCustomCommandPlugin())
    launcher.add_plugin(CmdApiPlugin())
    launcher.add_plugin(CmdSerialPlugin())
    launcher.add_plugin(CmdMousePlugin())
    launcher.add_plugin(CtxHistoryPlugin())
    launcher.add_plugin(OpenAIDallePlugin())
    launcher.add_plugin(OpenAIVisionPlugin())
    launcher.add_plugin(IdxLlamaIndexPlugin())
    launcher.add_plugin(MailerPlugin())
    launcher.add_plugin(CrontabPlugin())

    # register custom plugins
    plugins = kwargs.get('plugins', None)
    if isinstance(plugins, list):
        for plugin in plugins:
            launcher.add_plugin(plugin)

    # register base langchain and llama-index LLMs
    launcher.add_llm(OpenAILLM())
    launcher.add_llm(AzureOpenAILLM())
    launcher.add_llm(AnthropicLLM())
    launcher.add_llm(GoogleLLM())
    launcher.add_llm(HuggingFaceLLM())
    launcher.add_llm(HuggingFaceApiLLM())
    launcher.add_llm(LocalLLM())
    launcher.add_llm(OllamaLLM())
    launcher.add_llm(DeepseekApiLLM())

    # register custom langchain and llama-index LLMs
    llms = kwargs.get('llms', None)
    if isinstance(llms, list):
        for llm in llms:
            launcher.add_llm(llm)

    # register base vector store providers (llama-index)
    launcher.add_vector_store(ChromaProvider())
    launcher.add_vector_store(ElasticsearchProvider())
    launcher.add_vector_store(PinecodeProvider())
    launcher.add_vector_store(RedisProvider())
    launcher.add_vector_store(SimpleProvider())

    # register custom vector store providers (llama-index)
    vector_stores = kwargs.get('vector_stores', None)
    if isinstance(vector_stores, list):
        for store in vector_stores:
            launcher.add_vector_store(store)

    # register base llama-index agents
    launcher.add_agent(OpenAIAgent())
    launcher.add_agent(OpenAIAssistantAgent())
    launcher.add_agent(PlannerAgent())
    launcher.add_agent(ReactAgent())

    # register custom agents
    agents = kwargs.get('agents', None)
    if isinstance(agents, list):
        for agent in agents:
            launcher.add_agent(agent)

    # register base tools
    launcher.add_tool(IndexerTool())
    launcher.add_tool(MediaPlayerTool())
    launcher.add_tool(ImageViewerTool())
    launcher.add_tool(TextEditorTool())
    launcher.add_tool(AudioTranscriberTool())
    launcher.add_tool(CodeInterpreterTool())
    launcher.add_tool(HtmlCanvasTool())

    # register custom tools
    tools = kwargs.get('tools', None)
    if isinstance(tools, list):
        for tool in tools:
            launcher.add_tool(tool)

    # run the app
    launcher.run()
    

if __name__ == '__main__':
    run()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.25 01:00:00                  #
# ================================================== #
import json
import os
import requests

from .base import BaseProvider


class GoogleTextToSpeech(BaseProvider):
    def __init__(self, *args, **kwargs):
        """
        Google Text to Speech provider

        :param args: args
        :param kwargs: kwargs
        """
        super(GoogleTextToSpeech, self).__init__(*args, **kwargs)
        self.plugin = kwargs.get("plugin")
        self.id = "google"
        self.name = "Google TTS"

    def init_options(self):
        """Initialize options"""
        url_api = {
            "API Key": "https://console.cloud.google.com/apis/library/texttospeech.googleapis.com",
        }
        self.plugin.add_option(
            "google_api_key",
            type="text",
            value="",
            label="Google Cloud Text-to-speech API Key",
            tab="google",
            description="You can obtain your own API key at: "
                        "https://console.cloud.google.com/apis/library/texttospeech.googleapis.com",
            tooltip="Google Text-to-speech API Key",
            secret=True,
            persist=True,
            urls=url_api,
        )
        self.plugin.add_option(
            "google_voice",
            type="text",
            value="en-US-Neural2-J",
            label="Voice",
            tab="google",
            description="Specify voice",
            tooltip="Voice name",
            urls={
                "Voices": "https://cloud.google.com/text-to-speech/docs/voices"
            },
        )
        self.plugin.add_option(
            "google_lang",
            type="text",
            value="en-US",
            label="Language code",
            tab="google",
            description="Specify language code",
            tooltip="Language code",
            urls={
                "Languages": "https://cloud.google.com/speech-to-text/docs/speech-to-text-supported-languages"
            },
        )

    def speech(self, text: str) -> str:
        """
        Speech text to audio

        :param text: text to speech
        :return: path to generated audio file or None if audio playback is handled here
        """
        api_key = self.plugin.get_option_value("google_api_key")
        lang = self.plugin.get_option_value("google_lang")
        voice = self.plugin.get_option_value("google_voice")
        output_file = self.plugin.output_file
        path = os.path.join(
            self.plugin.window.core.config.path,
            output_file,
        )
        url = "https://texttospeech.googleapis.com/v1beta1/text:synthesize"
        data = {
            "input": {
                "text": text,
            },
            "voice": {
                "name": voice,
                "languageCode": lang,
            },
            "audioConfig": {
                "audioEncoding": "MP3",
            }
        }
        headers = {
            "content-type": "application/json",
            "X-Goog-Api-Key": api_key,
        }

        r = requests.post(url=url, json=data, headers=headers)
        content = json.loads(r.content)
        with open(path, "wb") as file:
            file.write(content.get("audioContent"))
        return path

    def is_configured(self) -> bool:
        """
        Check if provider is configured

        :return: True if configured, False otherwise
        """
        api_key = self.plugin.get_option_value("google_api_key")
        return api_key is not None and api_key != ""

    def get_config_message(self) -> str:
        """
        Return message to display when provider is not configured

        :return: message
        """
        api_key = self.plugin.get_option_value("google_api_key")
        if api_key is None or api_key == "":
            return "Google API KEY is not set. Please set it in plugin settings."




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

import re

from PySide6.QtCore import Qt, QObject, Signal, Slot, QEvent
from PySide6.QtWebChannel import QWebChannel
from PySide6.QtWebEngineCore import QWebEngineSettings, QWebEnginePage
from PySide6.QtWebEngineWidgets import QWebEngineView
from PySide6.QtGui import QAction, QIcon, QKeySequence

from pygpt_net.core.events import RenderEvent
from pygpt_net.item.ctx import CtxMeta
from pygpt_net.core.text.web_finder import WebFinder
from pygpt_net.utils import trans

import pygpt_net.icons_rc


class HtmlOutput(QWebEngineView):
    def __init__(self, window=None):
        """
        HTML output (Webkit)

        :param window: main window
        """
        super(HtmlOutput, self).__init__(window)
        self.window = window
        self.finder = WebFinder(window, self)
        self.loadFinished.connect(self.on_page_loaded)
        self.customContextMenuRequested.connect(self.on_context_menu)
        self.signals = WebEngineSignals()
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.plain = ""
        self.html_content = ""
        self.meta = None
        self.tab = None
        self.installEventFilter(self)

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def set_meta(self, meta: CtxMeta):
        """
        Assign ctx meta

        :param meta: context meta
        """
        self.meta = meta

    def set_plaintext(self, text: str):
        """
        Set plain text

        :param text: text
        """
        self.plain = text

    def set_html_content(self, html: str):
        """
        Set HTML content

        :param html: HTML content
        """
        self.html_content = html

        # set page HTML
        self.setHtml(html)

    def get_html_content(self) -> str:
        """
        Get HTML content

        :return: HTML content
        """
        return self.html_content

    def on_context_menu(self, position):
        """
        Context menu event

        :param position: position
        """
        menu = self.createStandardContextMenu()

        # remove defaults
        for action in menu.actions()[::-1]:
            menu.removeAction(action)

        selected_text = ""
        is_selection = self.page().hasSelection()
        if is_selection:
            selected_text = self.get_selected_text()

        if is_selection:
            # copy
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.copy'), self)
            action.triggered.connect(self.copy_selected_text)
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(self, selected_text)
            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.signals.save_as.emit(selected_text, 'txt')
            )
            menu.addAction(action)
        else:
            # select all
            action = QAction(QIcon(":/icons/copy.svg"), trans('action.select_all'), self)
            action.triggered.connect(self.select_all_text)
            menu.addAction(action)

            # save as (all) - html
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as') + " (html)", self)
            action.triggered.connect(
                lambda: self.signals.save_as.emit(re.sub(r'\n{2,}', '\n\n', self.html_content), 'html')
            )
            menu.addAction(action)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        action.setShortcut(QKeySequence("Ctrl+F"))
        menu.addAction(action)

        menu.exec_(self.mapToGlobal(position))

    def update_zoom(self):
        """Update zoom from config"""
        if self.window.core.config.has("zoom"):
            self.page().setZoomFactor(self.window.core.config.get("zoom"))

    def get_zoom_value(self) -> float:
        """
        Get zoom value

        :return: zoom value
        """
        return self.page().zoomFactor()

    def reset_current_content(self):
        """Reset current content"""
        self.plain = ""
        self.html_content = ""

    def update_current_content(self):
        """Update current content"""
        self.page().runJavaScript("document.getElementById('container').outerHTML", 0, self.set_plaintext)
        self.page().runJavaScript("document.documentElement.innerHTML", 0, self.set_html_content)

    def on_page_loaded(self, success):
        """
        On page loaded

        :param success: True if loaded successfully
        """
        if success:
            event = RenderEvent(RenderEvent.ON_PAGE_LOAD, {
                "meta": self.meta,
                "tab": self.tab,
            })
            self.window.dispatch(event)

    def get_selected_text(self) -> str:
        """
        Get selected text

        :return: selected text
        """
        return self.page().selectedText()

    def copy_selected_text(self):
        """Copy selected text"""
        self.page().triggerAction(QWebEnginePage.Copy)

    def select_all_text(self):
        """Select all text"""
        self.page().triggerAction(QWebEnginePage.SelectAll)

    def unselect_text(self):
        """Unselect text"""
        self.page().triggerAction(QWebEnginePage.Unselect)

    def find_open(self):
        """Open find dialog"""
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        """On content update"""
        self.finder.clear()  # clear finder

    def focusInEvent(self, e):
        """
        Focus in event

        :param e: focus event
        """
        super(HtmlOutput, self).focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if (event.type() == QEvent.ChildAdded and
                source is self and
                event.child().isWidgetType()):
            self._glwidget = event.child()
            self._glwidget.installEventFilter(self)
        elif (event.type() == event.Type.MouseButtonPress):
            if self.tab:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)


class CustomWebEnginePage(QWebEnginePage):
    """Custom WebEnginePage to handle web events"""
    def __init__(self, window, parent):
        super(CustomWebEnginePage, self).__init__(window)
        self.window = window
        self.parent = parent
        self.signals = WebEnginePageSignals()
        self.findTextFinished.connect(self.on_find_finished)
        self.contentsSizeChanged.connect(self.on_view_changed)
        self.selectionChanged.connect(self.on_selection_changed)
        self.settings().setAttribute(
            QWebEngineSettings.LocalContentCanAccessFileUrls, True
        )
        self.settings().setAttribute(
            QWebEngineSettings.LocalContentCanAccessRemoteUrls, True
        )
        self.settings().setFontFamily(QWebEngineSettings.StandardFont, 'Lato')
        self.settings().setFontFamily(QWebEngineSettings.FixedFont, 'Monaspace Neon')
        self.settings().setFontFamily(QWebEngineSettings.SerifFont, 'Monaspace Neon')
        if self.window.core.config.has("zoom"):
            self.setZoomFactor(self.window.core.config.get("zoom"))

        # bridge Python <> JavaScript
        self.bridge = Bridge(self.window)
        self.channel = QWebChannel(self)
        self.channel.registerObject("bridge", self.bridge)
        self.setWebChannel(self.channel)

    def on_find_finished(self, result):
        """
        On find text finished

        :param result: Find result
        """
        current = int(result.activeMatch())
        num = int(result.numberOfMatches())
        self.parent.finder.current_match_index = current
        self.parent.finder.matches = num
        self.parent.finder.on_find_finished()

    def on_view_changed(self):
        """On view changed"""
        zoom = self.zoomFactor()
        self.window.core.config.set("zoom", zoom)
        option = self.window.controller.settings.editor.get_option('zoom')
        option['value'] = zoom
        self.window.controller.config.apply(
            parent_id='config',
            key='zoom',
            option=option,
        )

    def on_selection_changed(self):
        """On selection changed"""
        pass

    def acceptNavigationRequest(self, url,  _type, isMainFrame):
        """
        On navigation (link click) event

        :param url: URL
        :param _type: Navigation type
        :param isMainFrame: True if main frame
        """
        if _type == QWebEnginePage.NavigationTypeLinkClicked:
            self.window.core.filesystem.url.handle(url)
            return False
        return super().acceptNavigationRequest(url,  _type, isMainFrame)

    def javaScriptConsoleMessage(self, level, message, line_number, source_id):
        """
        On JavaScript console message

        :param level: log level
        :param message: message
        :param line_number: line number
        :param source_id: source ID
        """
        self.signals.js_message.emit(line_number, message, source_id)  # handled in debug controller


class Bridge(QObject):
    """Bridge between Python and JavaScript"""
    def __init__(self, window):
        super(Bridge, self).__init__(window)
        self.window = window

    @Slot(str)
    def copy_text(self, text: str):
        """
        Copy text from web to clipboard

        :param text: text
        """
        self.window.controller.ctx.extra.copy_code_text(text)

    @Slot(int)
    def update_scroll_position(self, pos: int):
        """
        Update scroll position from web view

        :param pos: scroll position
        """
        self.window.controller.chat.render.scroll = pos


class WebEngineSignals(QObject):
    save_as = Signal(str, str)
    audio_read = Signal(str)


class WebEnginePageSignals(QObject):
    js_message = Signal(int, str, str)  # on Javascript message


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List

from PySide6.QtCore import Slot, QTimer

from pygpt_net.core.events import RenderEvent
from pygpt_net.core.render.base import BaseRenderer
from pygpt_net.core.render.markdown.renderer import Renderer as MarkdownRenderer
from pygpt_net.core.render.plain.renderer import Renderer as PlainTextRenderer
from pygpt_net.core.render.web.renderer import Renderer as WebRenderer
from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.text.utils import output_html2text, output_clean_html
from pygpt_net.item.ctx import CtxItem, CtxMeta


class Render:
    def __init__(self, window=None):
        """
        Render controller

        :param window: Window instance
        """
        self.window = window
        self.markdown_renderer = MarkdownRenderer(window)
        self.plaintext_renderer = PlainTextRenderer(window)
        self.web_renderer = WebRenderer(window)
        self.engine = None
        self.scroll = 0

    def setup(self):
        """Setup render"""
        self.engine = self.window.core.config.get("render.engine")

    def prepare(self):
        """Prepare render"""
        self.markdown_renderer.prepare()
        self.plaintext_renderer.prepare()
        self.web_renderer.prepare()

    def handle(self, event: RenderEvent):
        """
        Handle render event

        :param event: RenderEvent
        """
        name = event.name
        data = event.data
        if data is None:
            data = {}

        chunk = data.get("chunk", "")
        begin = data.get("begin", False)
        stream = data.get("stream", False)
        ctx = data.get("ctx")
        meta = data.get("meta")
        tool_data = data.get("tool_data")
        items = data.get("items")
        clear = data.get("clear", False)
        flush = data.get("flush", False)
        append = data.get("append", False)
        footer = data.get("footer", False)
        initialized = data.get("initialized", False)
        tab = data.get("tab")

        if name == RenderEvent.BEGIN:
            self.begin(meta, ctx, stream)
        elif name == RenderEvent.END:
            self.end(meta, ctx, stream)
        elif name == RenderEvent.RELOAD:
            self.reload()
        elif name == RenderEvent.RESET:
            self.reset(meta)
        elif name == RenderEvent.PREPARE:
            self.prepare()

        elif name == RenderEvent.STREAM_BEGIN:
            self.stream_begin(meta, ctx)
        elif name == RenderEvent.STREAM_APPEND:
            self.append_chunk(meta, ctx, chunk, begin)
        elif name == RenderEvent.STREAM_END:
            self.stream_end(meta, ctx)

        elif name == RenderEvent.ON_PAGE_LOAD:
            self.on_page_loaded(meta, tab)
        elif name ==RenderEvent.ON_THEME_CHANGE:
            self.on_theme_change()
        elif name == RenderEvent.ON_LOAD:
            self.on_load(meta)
        elif name == RenderEvent.ON_TS_ENABLE:
            self.on_enable_timestamp(live=initialized)
        elif name == RenderEvent.ON_TS_DISABLE:
            self.on_disable_timestamp(live=initialized)
        elif name == RenderEvent.ON_EDIT_ENABLE:
            self.on_enable_edit(live=initialized)
        elif name == RenderEvent.ON_EDIT_DISABLE:
            self.on_disable_edit(live=initialized)
        elif name == RenderEvent.ON_SWITCH:
            self.switch()

        elif name == RenderEvent.CLEAR_INPUT:
            self.clear_input()
        elif name == RenderEvent.CLEAR_OUTPUT:
            self.clear_output(meta)
        elif name == RenderEvent.CLEAR_ALL:
            self.clear_all()
        elif name == RenderEvent.CLEAR:
            self.clear(meta)

        elif name == RenderEvent.TOOL_UPDATE:
            self.tool_output_update(meta, tool_data)
        elif name == RenderEvent.TOOL_CLEAR:
            self.tool_output_clear(meta)
        elif name == RenderEvent.TOOL_BEGIN:
            self.tool_output_begin(meta)
        elif name == RenderEvent.TOOL_END:
            self.tool_output_end()

        elif name == RenderEvent.CTX_APPEND:
            self.append_context(meta, items, clear)
        elif name == RenderEvent.INPUT_APPEND:
            self.append_input(meta, ctx, flush, append)
        elif name == RenderEvent.OUTPUT_APPEND:
            self.append_output(meta, ctx)

        elif name == RenderEvent.EXTRA_APPEND:
            self.append_extra(meta, ctx, footer)
        elif name == RenderEvent.EXTRA_END:
            self.end_extra(meta, ctx)

        elif name == RenderEvent.ACTION_REGEN_SUBMIT:
            self.on_reply_submit(ctx)
        elif name == RenderEvent.ACTION_EDIT_SUBMIT:
            self.on_edit_submit(ctx)


    def get_pid(self, meta: CtxMeta) -> int:
        """
        Get PID for context meta

        :param meta: context meta
        :return: PID
        """
        return self.instance().get_pid(meta)

    def begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render begin

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().begin(meta, ctx, stream)
        self.update()

    def end(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end

        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().end(meta, ctx, stream)
        self.update()

    def end_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end extra
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if it is a stream
        """
        self.instance().end_extra(meta, ctx, stream)
        self.update()

    def stream_begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream begin
        
        :param meta: context meta
        :param ctx: context item
        """
        self.instance().stream_begin(meta, ctx)
        self.update()

    def stream_end(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream end
        
        :param meta: context meta
        :param ctx: context item
        """
        self.instance().stream_end(meta, ctx)
        self.update()

    def clear_output(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Clear current active output

        :param meta: context meta
        """
        self.instance().clear_output(meta) # TODO: get meta id on load
        self.update()

    def clear_input(self):
        """Clear input"""
        self.instance().clear_input()

    def on_load(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        On load (meta)

        :param meta: context meta
        """
        self.instance().on_load(meta)
        self.update()
        self.window.controller.ui.tabs.update_tooltip(meta.name)  # update tab tooltip

    def reset(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Reset current meta

        :param meta: Context meta
        """
        self.instance().reset(meta)  # TODO: get meta id on load
        self.update()

    def reload(self):
        """Reload current output"""
        self.instance().reload()  # TODO: or all outputs?
        self.update()

    def append_context(
            self,
            meta: CtxMeta,
            items: List[CtxItem],
            clear: bool = True
    ):
        """
        Append all context to output

        :param meta: context meta
        :param items: context items
        :param clear: True if clear all output before append
        """
        self.instance().append_context(meta, items, clear)
        self.update()

    def append_input(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            flush: bool = True,
            append: bool = False
    ):
        """
        Append text input to output
        
        :param meta: context meta
        :param ctx: context item
        :param flush: True if flush output
        :param append: True to force append node
        """
        self.instance().append_input(meta, ctx, flush=flush, append=append)
        self.update()

    def append_output(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Append text output to output
        
        :param meta: context meta
        :param ctx: context item
        """
        self.instance().append_output(meta, ctx)
        self.update()

    def append_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            footer: bool = False
    ):
        """
        Append extra data (images, files, etc.) to output
        
        :param meta: context meta
        :param ctx: context item
        :param footer: True if it is a footer
        """
        self.instance().append_extra(meta, ctx, footer)
        self.update()

    def append_chunk(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text_chunk: str,
            begin: bool = False
    ):
        """
        Append output stream chunk to output
        
        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the stream
        """
        self.instance().append_chunk(meta, ctx, text_chunk, begin)
        self.update()

    def on_enable_edit(self, live: bool = True):
        """
        On enable edit icons - global

        :param live: True if live update
        """
        self.instance().on_enable_edit(live)
        self.update()

    def on_disable_edit(self, live: bool = True):
        """
        On disable edit icons - global

        :param live: True if live update
        """
        self.instance().on_disable_edit(live)
        self.update()

    def on_enable_timestamp(self, live: bool = True):
        """
        On enable timestamp - global

        :param live: True if live update
        """
        self.instance().on_enable_timestamp(live)
        self.update()

    def on_disable_timestamp(self, live: bool = True):
        """
        On disable timestamp - global

        :param live: True if live update
        """
        self.instance().on_disable_timestamp(live)
        self.update()

    def remove_item(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        self.instance().remove_item(id)
        self.update()

    def remove_items_from(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        self.instance().remove_items_from(id)
        self.update()

    def on_edit_submit(self, ctx: CtxItem):
        """
        On edit submit

        :param ctx: context item
        """
        self.instance().on_edit_submit(ctx)
        self.update()

    def on_remove_submit(self, ctx: CtxItem):
        """
        On remove submit

        :param ctx: context item
        """
        self.instance().on_remove_submit(ctx)
        self.update()

    def on_reply_submit(self, ctx: CtxItem):
        """
        On regenerate submit

        :param ctx: context item
        """
        self.instance().on_reply_submit(ctx)
        self.update()

    def on_page_loaded(
            self,
            meta: CtxMeta,
            tab: Optional[Tab] = None
    ):
        """
        On page loaded callback

        :param meta: context meta
        :param tab: Tab
        """
        self.instance().on_page_loaded(meta, tab)
        self.update()

    def on_theme_change(self):
        """On theme change - global"""
        if self.get_engine() == "web":
            self.web_renderer.on_theme_change()
        elif self.get_engine() == "legacy":
            self.markdown_renderer.on_theme_change()
        self.update()

    def get_scroll_position(self) -> int:
        """
        Get scroll position - active

        :return: scroll position
        """
        return self.instance().get_scroll_position()

    def set_scroll_position(self, position: int):
        """
        Set scroll position - active

        :param position: scroll position
        """
        self.instance().set_scroll_position(position)
        self.update()

    def update(self):
        """On update - active"""
        for pin in self.window.ui.nodes['output']:
            self.window.ui.nodes['output'][pin].on_update()

    def clear(self, meta: CtxMeta):
        """
        Clear renderer

        :param meta: ctx meta instance
        """
        self.reset(meta)
        self.clear_output(meta)

    def clear_all(self):
        """Clear all"""
        self.instance().clear_all()
        self.update()

    def tool_output_append(
            self,
            meta: CtxMeta,
            content: str
    ):
        """
        Add tool output (append)

        :param meta: context meta
        :param content: content
        """
        self.instance().tool_output_append(meta, content)
        self.update()

    def tool_output_update(
            self,
            meta: CtxMeta,
            content: str
    ):
        """
        Replace tool output

        :param meta: context meta
        :param content: content
        """
        self.instance().tool_output_update(meta, content)
        self.update()

    def tool_output_clear(self, meta: CtxMeta):
        """
        Clear tool output

        :param meta: context meta
        """
        self.instance().tool_output_clear(meta)
        self.update()

    def tool_output_begin(self, meta: CtxMeta):
        """
        Begin tool output

        :param meta: context meta
        """
        self.instance().tool_output_begin(meta)
        self.update()

    def tool_output_end(self):
        """
        End tool output

        :param meta: context meta
        """
        self.instance().tool_output_end()
        self.update()

    def get_engine(self) -> str:
        """
        Get current render engine ID

        :return: engine name
        """
        return self.engine

    def switch(self):
        """
        Switch renderer (markdown/web <==> plain text) - active, TODO: remove from settings, leave only checkbox
        """
        plain = self.window.core.config.get('render.plain')
        if plain:
            self.window.controller.theme.markdown.clear()
            for pid in self.window.ui.nodes['output_plain']:
                if self.window.ui.nodes['output_plain'][pid] is not None:
                    self.window.ui.nodes['output'][pid].setVisible(False)
                    self.window.ui.nodes['output_plain'][pid].setVisible(True)
        else:
            self.window.controller.ctx.refresh()  # TODO: move to on_switch
            self.window.controller.theme.markdown.update(force=True)
            for pid in self.window.ui.nodes['output']:
                if self.window.ui.nodes['output'][pid] is not None:
                    self.window.ui.nodes['output'][pid].setVisible(True)
                    self.window.ui.nodes['output_plain'][pid].setVisible(False)

    def instance(self) -> BaseRenderer:
        """
        Get instance of current renderer

        :return: Renderer instance
        """
        # get selected renderer
        if self.window.core.config.get('render.plain'):
            return self.plaintext_renderer
        else:
            if self.engine == "web":
                return self.web_renderer
            else:
                return self.markdown_renderer

    @Slot(str, str)
    def handle_save_as(
            self,
            text: str,
            type: str = 'txt'
    ):
        """
        Handle save as signal  # TODO: move to another class

        :param text: Data to save
        :param type: File type
        """
        if type == 'html':
            text = output_clean_html(text)
        else:
            text = output_html2text(text)
        # fix: QTimer required here to prevent crash if signal emitted from WebEngine window
        QTimer.singleShot(0, lambda: self.window.controller.chat.common.save_text(text, type))

    @Slot(str)
    def handle_audio_read(self, text: str):
        """
        Handle audio read signal

        :param text: Text to read  # TODO: move to another class
        """
        self.window.controller.audio.read_text(text)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

from typing import Any

from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_ASSISTANT,
    MODE_IMAGE,
)
from pygpt_net.core.events import Event, AppEvent, RenderEvent, KernelEvent
from pygpt_net.item.ctx import CtxItem


class Output:
    def __init__(self, window=None):
        """
        Output controller

        :param window: Window instance
        """
        self.window = window
        self.not_stream_modes = [
            MODE_ASSISTANT,
            MODE_IMAGE,
        ]

    def handle(
            self,
            ctx: CtxItem,
            mode: str,
            stream_mode: bool = False
    ):
        """
        Handle response from LLM

        :param ctx: CtxItem
        :param mode: mode (global)
        :param stream_mode: stream mode
        """
        self.window.stateChanged.emit(self.window.STATE_BUSY)

        # if stream mode then append chunk by chunk
        stream_enabled = self.window.core.config.get("stream")  # global stream enabled
        if stream_mode:  # local stream enabled
            if mode not in self.not_stream_modes:
                self.window.controller.chat.stream.append(ctx)

        # check if tool calls detected
        if ctx.tool_calls:
            # if not internal commands in a text body then append tool calls as commands (prevent double commands)
            if not self.window.core.command.has_cmds(ctx.output):
                self.window.core.command.append_tool_calls(ctx)  # append tool calls as commands
                if not isinstance(ctx.extra, dict):
                    ctx.extra = {}
                ctx.extra["tool_calls"] = ctx.tool_calls
                stream_mode = False  # disable stream mode, show tool calls at the end
                self.log("Tool call received...")
            else:  # prevent execute twice
                self.log("Ignoring tool call because command received...")

        # agent mode
        if mode == MODE_AGENT:
            self.window.controller.agent.legacy.on_ctx_after(ctx)

        # event: context after
        event = Event(Event.CTX_AFTER)
        event.ctx = ctx
        self.window.dispatch(event)

        self.log("Appending output to chat window...")

        # only append output if not in stream mode, TODO: plugin output add
        if not stream_mode:
            if stream_enabled:  # use global stream settings here to persist previously added input
                data = {
                    "meta": ctx.meta,
                    "ctx": ctx,
                    "flush": True,
                    "append": True,
                }
                event = RenderEvent(RenderEvent.INPUT_APPEND, data)
                self.window.dispatch(event)

            data = {
                "meta": ctx.meta,
                "ctx": ctx,
            }
            event = RenderEvent(RenderEvent.OUTPUT_APPEND, data)
            self.window.dispatch(event)

            data = {
                "meta": ctx.meta,
                "ctx": ctx,
                "footer": True,
            }
            event = RenderEvent(RenderEvent.EXTRA_APPEND, data)
            self.window.dispatch(event)  # + icons

        self.handle_complete(ctx)

    def handle_complete(self, ctx: CtxItem):
        """
        Handle completed context

        :param ctx: CtxItem
        """
        mode = self.window.core.config.get('mode')
        self.window.core.ctx.post_update(mode)  # post update context, store last mode, etc.
        self.window.core.ctx.store()
        self.window.controller.ctx.update_ctx()  # update current ctx info

        # update response tokens
        self.window.controller.chat.common.show_response_tokens(ctx)

        # handle audio output
        self.window.controller.chat.audio.handle_output(ctx)

        # store to history
        if self.window.core.config.get('store_history'):
            self.window.core.history.append(ctx, "output")

        if self.window.controller.chat.common.can_unlock(ctx):
            if not self.window.controller.kernel.stopped():
                self.window.controller.chat.common.unlock_input()  # unlock input

        # reset state to: idle
        self.window.dispatch(KernelEvent(KernelEvent.STATE_IDLE, {
            "id": "chat",
        }))

    def post_handle(
            self,
            ctx: CtxItem,
            mode: str,
            stream: bool = False,
            reply: bool = False,
            internal: bool = False
    ):
        """
        Post handle results

        :param ctx: CtxItem
        :param mode: mode (global)
        :param stream: stream mode
        :param reply: is reply
        :param internal: is internal
        """
        # if commands enabled: post-execute commands (not assistant mode)
        if mode != MODE_ASSISTANT:
            ctx.clear_reply()  # reset results

            # extract expert mentions and handle experts reply
            num_calls = self.window.controller.agent.experts.handle(ctx)
            if num_calls == 0:
                # handle commands only if no expert calls in queue
                self.window.controller.chat.command.handle(ctx)

            ctx.from_previous()  # append previous result again before save
            self.window.core.ctx.update_item(ctx)  # update ctx in DB

        # render: end
        if ctx.sub_calls == 0:  # if no experts called
            data = {
                "meta": ctx.meta,
                "ctx": ctx,
                "stream": stream,
            }
            event = RenderEvent(RenderEvent.END, data)
            self.window.dispatch(event)

        # don't unlock input and leave stop btn if assistant mode or if agent/autonomous is enabled
        # send btn will be unlocked in agent mode on stop
        if self.window.controller.chat.common.can_unlock(ctx):
            if not self.window.controller.kernel.stopped():
                self.window.controller.chat.common.unlock_input()  # unlock input

        # handle ctx name (generate title from summary if not initialized)
        if not reply and not internal:  # don't call if reply or internal mode
            if self.window.core.config.get('ctx.auto_summary'):
                self.log("Calling for prepare context name...")
                self.window.controller.ctx.prepare_name(ctx)  # async

    def handle_end(
            self,
            ctx: CtxItem,
            mode: str
    ):
        """
        Handle context end (finish output)

        :param ctx: CtxItem
        :param mode: mode
        """
        # clear attachments after send, only if attachments has been provided before send
        auto_clear = self.window.core.config.get('attachments_send_clear')
        if self.window.controller.attachment.clear_allowed(ctx):
            if auto_clear and not self.window.controller.attachment.is_locked():
                self.window.controller.attachment.clear(force=True, auto=True)
                self.window.controller.attachment.update()
                self.log("Attachments cleared.")  # log

        if self.window.core.config.get("log.ctx"):
            self.log("Context: END: {}".format(ctx))
        else:
            self.log("Context: END.")

        # agent mode
        if mode == MODE_AGENT:
            agent_iterations = int(self.window.core.config.get("agent.iterations"))
            self.log("Agent: ctx end, iterations: {}".format(agent_iterations))
            self.window.controller.agent.legacy.on_ctx_end(
                ctx,
                iterations=agent_iterations,
            )

        # event: context end
        event = Event(Event.CTX_END)
        event.ctx = ctx
        self.window.dispatch(event)
        self.window.controller.ui.update_tokens()  # update UI
        self.window.controller.chat.input.generating = False  # unlock

        if (mode not in self.window.controller.chat.input.no_ctx_idx_modes
                and not self.window.controller.agent.legacy.enabled()):
            self.window.controller.idx.on_ctx_end(ctx, mode=mode)  # update ctx DB index
            # disabled in agent mode here to prevent loops, handled in agent flow internally if agent mode

        self.log("End.")
        self.window.dispatch(AppEvent(AppEvent.CTX_END))  # app event

        # restore state to idle if no errors
        if self.window.state != self.window.STATE_ERROR:
            self.window.stateChanged.emit(self.window.STATE_IDLE)

        if mode != MODE_ASSISTANT:
            self.window.controller.kernel.stack.handle()  # handle reply
            event = RenderEvent(RenderEvent.RELOAD)
            self.window.dispatch(event)  # reload chat window

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.controller.chat.log(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Dict, Any


class BaseAgent:
    def __init__(self, *args, **kwargs):
        self.id = ""
        self.mode = "step"  # step|plan

    def get_mode(self) -> str:
        """
        Return Agent mode

        :return: Agent mode
        """
        return self.mode

    def get_agent(self, window, kwargs: Dict[str, Any]):
        """
        Return Agent provider instance

        :param window: window instance
        :param kwargs: keyword arguments
        :return: Agent provider instance
        """
        pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 03:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "voice_control"
        self.name = "Voice Control (inline)"
        self.type = [
            "cmd.inline",
            "audio.control",
        ]
        self.description = "Provide voice control command execution within a conversation."
        self.prefix = "Voice"
        self.input_text = None
        self.allowed_cmds = [
            "voice_cmd",
        ]
        self.order = 100
        self.use_locale = True
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name in [
            Event.CMD_SYNTAX_INLINE,  # inline is allowed
            Event.CMD_SYNTAX,
        ]:
            self.cmd_syntax(data)

        elif name in [
            Event.CMD_INLINE,  # inline is allowed
            Event.CMD_EXECUTE,
        ]:
            self.cmd(
                ctx,
                data['commands'],
            )

        elif name in [
            Event.ENABLE,
            Event.DISABLE,
        ]:
            if name == Event.ENABLE:
                if data["value"] == self.id:
                    self.window.controller.plugins.enable("audio_input")

    def prepare_cmd_list(self) -> dict:
        """
        Prepare command list

        :return: command list
        """
        magic_prefix = self.get_option_value("cmd_prefix")
        instruction = self.window.core.access.voice.get_inline_prompt(magic_prefix).strip()
        cmd_syntax = {
            "cmd": "voice_cmd",
            "instruction": instruction,
            "params": [
                {
                    "name": "action",
                    "type": "enum",
                    "required": True,
                    "description": "Voice action",
                    "enum": {
                        "action": self.window.core.access.voice.get_commands(),  # voice cmds list
                    },
                },
                {
                    "name": "args",
                    "type": "text",
                    "required": False,
                    "description": "Voice action arguments",
                }
            ],
            "enabled": True,  # enabled
        }
        return cmd_syntax

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        data['cmd'].append(self.prepare_cmd_list())  # append command

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Events: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds:
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        # prepare voice commands
        voice_commands = []
        available_commands = self.window.core.access.voice.get_commands()
        for cmd in my_commands:
            if cmd["cmd"] == "voice_cmd":
                if "action" not in cmd["params"]:
                    return
                action = cmd["params"]["action"]
                params = ""
                if "args" in cmd["params"]:
                    params = cmd["params"]["args"]
                if action not in available_commands:
                    voice_commands.append({
                        "cmd": "unrecognized",
                        "params": "",
                    })
                    continue
                voice_commands.append({
                    "cmd": action,
                    "params": params,
                })

        # execute voice commands
        self.window.controller.access.voice.handle_commands(voice_commands)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 03:00:00                  #
# ================================================== #

import json
import ssl

from urllib.parse import urlencode
from urllib.request import Request, urlopen

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_api"
        self.name = "API calls"
        self.description = "Provides the ability to make external API calls"
        self.prefix = "API"
        self.order = 100
        self.use_locale = True
        self.worker = None
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: event args
        :param kwargs: event kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name == Event.CMD_SYNTAX:
            self.cmd_syntax(data)

        elif name == Event.CMD_EXECUTE:
            self.cmd(
                ctx,
                data['commands'],
            )

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for item in self.get_option_value("cmds"):
            if not item["enabled"]:
                continue
            cmd = {
                "cmd": item["name"],
                "instruction": item["instruction"],
                "params": [],
            }
            # GET
            if item["get_params"].strip() != "":
                params = self.extract_params(item["get_params"])
                if len(params) > 0:
                    for param in params:
                        cmd["params"].append(param)
            # POST
            if item["post_params"].strip() != "":
                params = self.extract_params(item["post_params"])
                if len(params) > 0:
                    for param in params:
                        cmd["params"].append(param)

            data['cmd'].append(cmd)

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Event: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            for my_cmd in self.get_option_value("cmds"):
                if not my_cmd["enabled"]:
                    continue
                if my_cmd["name"] == item["cmd"]:
                    is_cmd = True
                    my_commands.append(item)

        if not is_cmd:
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        try:
            worker = Worker()
            worker.from_defaults(self)
            worker.cmds = my_commands
            worker.ctx = ctx

            if not self.is_async(ctx):
                worker.run()
                return
            worker.run_async()

        except Exception as e:
            self.error(e)

    def get_item(self, name: str) -> dict:
        """
        Get API call by name

        :param name: call name
        :return: call dict
        """
        for item in self.get_option_value("cmds"):
            if item["name"] == name:
                return item
        return {}

    def extract_params(self, text: str) -> list:
        """
        Extract params from params string

        :param text: text
        :return: params list
        """
        params = []
        if text is None or text == "":
            return params
        params_list = text.split(",")
        for param in params_list:
            param = param.strip()
            if param == "":
                continue
            params.append({
                "name": param,
                "type": "str",
                "description": param,
            })
        return params

    def call_get(self, url: str, extra_headers: dict = None) -> bytes:
        """
        Call API with GET method

        :param url: URL
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # log API call details
        log_data = {
            "url": url,
            "headers": headers,
            "type": "GET",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data

    def call_post(self, url: str, data: dict, extra_headers: dict = None) -> bytes:
        """
        Call API with POST method

        :param url: URL
        :param data: POST data dict
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # encode data
        data = urlencode(data).encode()

        # log API call details
        log_data = {
            "url": url,
            "data": data,
            "headers": headers,
            "type": "POST",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            data=data,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data

    def call_post_json(self, url: str, data: dict, extra_headers: dict = None) -> bytes:
        """
        Call API with POST method (JSON)

        :param url: URL
        :param data: POST data dict
        :param extra_headers: extra headers
        :return: response data
        """
        headers = {
            'User-Agent': self.get_option_value("user_agent"),
            'Content-Type': 'application/json',
        }
        if extra_headers is not None:
            headers.update(extra_headers)

        # encode data
        data = json.dumps(data).encode()

        # log API call details
        log_data = {
            "url": url,
            "data": data,
            "headers": headers,
            "type": "POST JSON",
        }
        self.log(str(log_data))

        # prepare request
        req = Request(
            url=url,
            data=data,
            headers=headers,
        )
        if self.get_option_value('disable_ssl'):
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            data = urlopen(
                req,
                context=context,
                timeout=self.get_option_value('timeout'),
            ).read()
        else:
            data = urlopen(
                req,
                timeout=self.get_option_value('timeout'),
            ).read()
        return data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.09 00:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QTextBrowser
from PySide6.QtGui import QAction, QIcon, QTextOption, QKeySequence

from pygpt_net.core.text.finder import Finder
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class ChatOutput(QTextBrowser):
    def __init__(self, window=None):
        """
        Chat output

        :param window: main window
        """
        super(ChatOutput, self).__init__(window)
        self.window = window
        self.finder = Finder(window, self)
        self.setReadOnly(True)
        self.setAcceptRichText(False)
        self.setStyleSheet(self.window.controller.theme.style('font.chat.output'))
        self.value = self.window.core.config.get('font_size')
        self.max_font_size = 42
        self.min_font_size = 8
        self.setOpenExternalLinks(False)
        self.setOpenLinks(False)
        self.anchorClicked.connect(self.open_external_link)
        self.setWordWrapMode(QTextOption.WordWrap)
        self.tab = None

    def set_tab(self, tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def open_external_link(self, url):
        """
        Handle external link

        :param url: url
        """
        self.window.core.filesystem.url.handle(url)

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: Event
        """
        menu = self.createStandardContextMenu()
        selected_text = self.textCursor().selectedText()

        if selected_text:
            # plain text
            plain_text = self.textCursor().selection().toPlainText()

            # audio read
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            # copy to
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(self, selected_text)
            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text)
            )
            menu.addAction(action)
        else:
            # save as (all)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText())
            )
            menu.addAction(action)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        action.setShortcut(QKeySequence("Ctrl+F"))
        menu.addAction(action)

        menu.exec_(event.globalPos())

    def audio_read_selection(self):
        """Read selected text (audio)"""
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def find_open(self):
        """Open find dialog"""
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        """On content update"""
        self.finder.clear()  # clear finder

    def keyPressEvent(self, e):
        """
        Key press event

        :param e: Event
        """
        if e.key() == Qt.Key_F and e.modifiers() & Qt.ControlModifier:
            self.find_open()  # open find dialog
        else:
            super(ChatOutput, self).keyPressEvent(e)

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            self.window.core.config.data['font_size'] = self.value
            self.window.core.config.save()
            option = self.window.controller.settings.editor.get_option('font_size')
            option['value'] = self.value
            self.window.controller.config.apply(
                parent_id='config', 
                key='font_size', 
                option=option,
            )
            self.window.controller.ui.update_font_size()
            event.accept()
        else:
            super(ChatOutput, self).wheelEvent(event)

    def focusInEvent(self, e):
        """
        Focus in event

        :param e: focus event
        """
        super(ChatOutput, self).focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_EXPERT,
)

class Experts:
    def __init__(self, window=None):
        """
        Experts controller

        :param window: Window instance
        """
        self.window = window

    def refresh(self):
        """Refresh presets"""
        pass

    def get_current_agent_id(self) -> str:
        """
        Get current agent UUID

        :return: agent UUID
        """
        return self.window.controller.presets.editor.current

    def is_active(self, uuid: str) -> bool:
        """
        Check if expert is active in current agent

        :param uuid: expert uuid
        :return: bool
        """
        agent_uuid = self.get_current_agent_id()
        if agent_uuid is None:
            return False
        agent = self.window.core.presets.get_by_uuid(agent_uuid)
        if agent is None:
            return False
        return uuid in agent.experts

    def update_list(self):
        """Update presets list"""
        items = self.window.core.presets.get_by_mode(MODE_EXPERT)
        items_data = {}
        for item in items:
            if item.startswith("current.") or self.is_active(items[item].uuid):
                continue
            items_data[item] = items[item]
        self.window.ui.nodes['preset.editor.experts'].update_available(items_data)  # update available
        current_agent = self.get_current_agent_id()
        if current_agent is not None:
            agent = self.window.core.presets.get_by_uuid(current_agent)
            if agent is None:
                return
            experts = {}
            for expert_uuid in agent.experts:
                expert = self.window.core.presets.get_by_uuid(expert_uuid)
                if expert is None:
                    continue
                experts[expert_uuid] = expert
            self.window.ui.nodes['preset.editor.experts'].update_selected(experts)  # update selected

    def change_available(self):
        """Change selected expert"""
        pass

    def change_selected(self):
        """Change selected expert"""
        pass

    def get_current_available(self) -> str:
        """
        Get current available expert

        :return: expert uuid
        """
        val = self.window.ui.nodes["preset.experts.available"].selectionModel().currentIndex()
        idx = val.row()
        return self.get_available_by_idx(idx)

    def get_current_selected(self) -> str:
        """
        Get current selected expert

        :return: expert uuid
        """
        val = self.window.ui.nodes["preset.experts.selected"].selectionModel().currentIndex()
        idx = val.row()
        return self.get_selected_by_idx(idx)

    def get_available_by_idx(self, idx):
        """
        Get available expert by index

        :param idx: index
        """
        presets = self.window.core.presets.get_by_mode(MODE_EXPERT)
        i = 0
        for preset in presets:
            if preset.startswith("current.") or self.is_active(presets[preset].uuid):
                continue
            if i == idx:
                return presets[preset].uuid
            i += 1

    def get_selected_by_idx(self, idx):
        """
        Get selected expert by index

        :param idx: index
        """
        agent_uuid = self.get_current_agent_id()
        agent = self.window.core.presets.get_by_uuid(agent_uuid)
        if agent is None:
            return None
        i = 0
        for expert_uuid in agent.experts:
            if i == idx:
                return expert_uuid
            i += 1

    def add_expert(self):
        """Add expert"""
        agent_uuid = self.get_current_agent_id()
        if agent_uuid is None or not self.window.core.presets.exists_uuid(agent_uuid):
            self.window.controller.presets.editor.save(close=False)
            return
        expert_uuid = self.get_current_available()
        if expert_uuid is None:
            return
        self.window.core.presets.add_expert(agent_uuid, expert_uuid)
        self.update_list()

    def remove_expert(self):
        """Remove expert"""
        agent_uuid = self.get_current_agent_id()
        if agent_uuid is None or not self.window.core.presets.exists_uuid(agent_uuid):
            self.window.controller.presets.editor.save(close=False)
            return
        expert_uuid = self.get_current_selected()
        if expert_uuid is None:
            return
        self.window.core.presets.remove_expert(agent_uuid, expert_uuid)
        self.update_list()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 04:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        plugin.add_option(
            "smtp_host",
            type="text",
            value="",
            label="SMTP Host",
            persist=True,
            tab="smtp",
        )
        plugin.add_option(
            "smtp_port_inbox",
            type="int",
            value=995,
            label="SMTP Port (Inbox)",
            persist=True,
            tab="smtp",
        )
        plugin.add_option(
            "smtp_port_outbox",
            type="int",
            value=465,
            label="SMTP Port (Outbox)",
            persist=True,
            tab="smtp",
        )
        plugin.add_option(
            "smtp_user",
            type="text",
            value="",
            label="SMTP User",
            persist=True,
            tab="smtp",
        )
        plugin.add_option(
            "smtp_password",
            type="text",
            value="",
            label="SMTP Password",
            secret=True,
            persist=True,
            tab="smtp",
        )
        plugin.add_option(
            "from_email",
            type="text",
            value="",
            label="From (email)",
        )
        plugin.add_cmd(
            "send_email",
            instruction="Send email to specified address",
            params=[
                {
                    "name": "recipient",
                    "type": "str",
                    "description": "Recipient email address",
                    "required": True,
                },
                {
                    "name": "subject",
                    "type": "str",
                    "description": "Email subject",
                    "required": True,
                },
                {
                    "name": "message",
                    "type": "str",
                    "description": "Email message",
                    "required": True,
                },
            ],
            enabled=True,
            description="Allows sending emails",
            tab="general",
        )
        plugin.add_cmd(
            "get_emails",
            instruction="Retrieve emails from the inbox. Use this to access emails from the inbox when the user "
                        "requests fetching emails from their server.",
            params=[
                {
                    "name": "limit",
                    "type": "int",
                    "description": "Limit of emails to receive, default: 10",
                    "required": True,
                },
                {
                    "name": "offset",
                    "type": "int",
                    "description": "Offset of emails to receive, default: 0",
                    "required": False,
                },
                {
                    "name": "order",
                    "type": "text",
                    "description": "Sort order, default: desc, options: asc, desc",
                    "required": False,
                },
            ],
            enabled=True,
            description="Allows receiving emails",
            tab="general",
        )
        plugin.add_cmd(
            "get_email_body",
            instruction="Retrieve the complete email content from the inbox. Use this to access the message body. "
                        "To obtain email IDs, use the 'get_emails' function first.",
            params=[
                {
                    "name": "id",
                    "type": "int",
                    "description": "Email ID (index)",
                    "required": True,
                },
                {
                    "name": "format",
                    "type": "text",
                    "description": "Display format, default: text, options: text, html",
                    "required": False,
                },
            ],
            enabled=True,
            description="Allows receiving email body",
            tab="general",
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List

from llama_index.core.tools import FunctionTool

from pygpt_net.item.ctx import CtxItem


class Evaluation:
    def __init__(self, window=None):
        self.window = window
        self.prompt = """
        Please review the result below to determine if the agent's response is satisfactory and if the assigned 
        task was completed correctly. Evaluate the quality and accuracy of the response, as well as the successful 
        completion of the task, using a percentage scale from 0% to 100%. Use the tool provided to send feedback to the agent, 
        including instructions addressed directly to him on how to improve the previous result, along with a 
        numerical rating. The instructions should be prepared in the language used by the user.
        
        ## Tool for sending feedback:
        
        - send_feedback
        
        ## When creating an instruction, please use the following format:
        
        ```
        Please extend your response by including the following:
        
        1. ...
        2. ...
        ```
        
        ## Content to evaluate:
        
        MAIN TASK:
        
        ```
        
        {task}
        
        ```
        
        LAST USER INPUT:
        
        ```
        
        {input}
        
        ```
        
        AGENT RESPONSE:
        
        ```
        
        {output}
        
        ```
        
        ## Additional rules:
        
        - ALWAYS provide the instruction for the agent in the language used by the user in main task description.
        - Do not repeat the suggested improvements if they have already been correctly included in the agent's response.
        """

    def get_last_user_input(self, history: List[CtxItem]) -> str:
        """
        Get the last user input from the history

        :param history: ctx items
        :return: last user input
        """
        input = ""
        use_prev = self.window.core.config.get("agent.llama.append_eval", False)
        for ctx in history:
            if ctx.extra is not None and "agent_input" in ctx.extra:
                if not use_prev and "agent_evaluate" in ctx.extra:  # exclude evaluation inputs
                        continue
                if ctx.input:
                    input = ctx.input
        return input

    def get_main_task(self, history: List[CtxItem]) -> str:
        """
        Get the main task from the history

        :param history: ctx items
        :return: main task
        """
        first = history[0]
        task = ""
        if first.extra is not None and "agent_input" in first.extra:
            task = first.input
        return task

    def get_final_response(self, history: List[CtxItem]) -> str:
        """
        Get the final response from the agent

        :param history: ctx items
        :return: final response
        """
        output = ""
        for ctx in history:
            if ctx.extra is not None and "agent_finish" in ctx.extra:
                if ctx.output:
                    output = ctx.output
        return output

    def get_prompt(self, history: List[CtxItem]) -> str:
        """
        Return the evaluation prompt

        :param history:
        :return: evaluation prompt
        """
        prompt = self.window.core.config.get("prompt.agent.llama.eval")
        main_task = self.get_main_task(history)
        last_input = self.get_last_user_input(history)
        final_response = self.get_final_response(history)
        return prompt.format(
            task=main_task,
            input=last_input,
            output=final_response
        )

    def get_tools(self) -> List[FunctionTool]:
        """
        Get the tools for evaluating the result

        :return: list of tools
        """
        def send_feedback(instructions: str, rating_percent: int) -> str:
            """Send feedback with evaluation result"""
            self.handle_evaluation(instructions, rating_percent)
            return "OK. Feedback has been sent."

        tool = FunctionTool.from_defaults(fn=send_feedback)
        tools = []
        tools.append(tool)
        return tools

    def handle_evaluation(
            self,
            instruction: str,
            score: int
    ):
        """
        Update the evaluation values of the agent response

        :param instruction: instruction
        :param score: score
        """
        self.window.core.agents.runner.next_instruction = instruction
        self.window.core.agents.runner.prev_score = score



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional

from .base import BaseEvent

class AppEvent(BaseEvent):
    """Events dispatched by application"""
    APP_STARTED = "app.started"
    CTX_CREATED = "ctx.created"
    CTX_END = "ctx.end"
    CTX_SELECTED = "ctx.selected"
    CTX_ATTACHMENTS_CLEAR = "ctx.attachments.clear"
    CAMERA_ENABLED = "camera.enabled"
    CAMERA_DISABLED = "camera.disabled"
    CAMERA_CAPTURED = "camera.captured"
    INPUT_ERROR = "input.error"
    INPUT_SENT = "input.sent"
    INPUT_CALL = "input.call"
    INPUT_STOPPED = "input.stopped"
    INPUT_VOICE_LISTEN_STARTED = "input.voice.listen.started"
    INPUT_VOICE_LISTEN_STOPPED = "input.voice.listen.stopped"
    MODE_SELECTED = "mode.selected"
    MODEL_SELECTED = "model.selected"
    PRESET_SELECTED = "preset.selected"
    TAB_SELECTED = "tab.switch"
    VOICE_CONTROL_TOGGLE = "voice.control.toggle"
    VOICE_CONTROL_STARTED = "voice.control.started"
    VOICE_CONTROL_STOPPED = "voice.control.stopped"
    VOICE_CONTROL_SENT = "voice.control.sent"
    VOICE_CONTROL_UNRECOGNIZED = "voice.control.unrecognized"

    def __init__(
            self,
            name: Optional[str] = None,
            data: Optional[dict] = None,
    ):
        """
        Event object class

        :param name: event name
        :param data: event data
        """
        super(AppEvent, self).__init__(name, data)
        self.id = "AppEvent"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QDialog, QLabel, QHBoxLayout, QVBoxLayout, QPushButton, QScrollArea, QWidget, QSizePolicy

from pygpt_net.ui.widget.element.group import QHLine
from pygpt_net.ui.widget.element.labels import HelpLabel, UrlLabel, IconLabel
from pygpt_net.ui.widget.option.combo import OptionCombo
from pygpt_net.utils import trans
from pygpt_net.ui.widget.textarea.url import UrlInput


class UrlDialog(QDialog):
    def __init__(self, window=None, id=None):
        """
        Url dialog

        :param window: main window
        :param id: info window id
        """
        super(UrlDialog, self).__init__(window)
        self.window = window
        self.id = id
        self.current = None
        #self.input = UrlInput(window, id)
        #self.input.setMinimumWidth(400)
        self.initialized = False
        self.params_scroll = None

    def init(self):
        """Initialize dialog"""
        if self.initialized:
            return

        self.window.ui.nodes['dialog.url.btn.update'] = QPushButton(trans('dialog.url.update'))
        self.window.ui.nodes['dialog.url.btn.update'].clicked.connect(
            lambda: self.window.controller.attachment.attach_url()
        )

        self.window.ui.nodes['dialog.url.btn.dismiss'] = QPushButton(trans('dialog.url.dismiss'))
        self.window.ui.nodes['dialog.url.btn.dismiss'].clicked.connect(
            lambda: self.window.controller.dialogs.confirm.dismiss_url())

        bottom = QHBoxLayout()
        bottom.addWidget(self.window.ui.nodes['dialog.url.btn.dismiss'])
        bottom.addWidget(self.window.ui.nodes['dialog.url.btn.update'])

        #self.window.ui.nodes['dialog.url.label'] = QLabel(trans("dialog.url.title"))
        #self.window.ui.nodes['dialog.url.tip'] = HelpLabel(trans("dialog.url.tip"))

        # -----------------

        loaders = self.window.controller.config.placeholder.apply_by_id("llama_index_loaders_web")
        loaders_list = []
        for loader in loaders:
            k = list(loader.keys())[0]
            key = k.replace("web_", "")
            value = loader[k]
            loaders_list.append({
                key: value,
            })

        self.window.ui.nodes["dialog.url.loader"] = OptionCombo(
            self.window,
            "dialog.url",
            "web.loader",
            {
                "label": trans("tool.indexer.tab.web.loader"),
                "keys": loaders_list,
                "value": "webpage",
            }
        )

        self.window.ui.nodes["dialog.url.loader"].layout.setContentsMargins(0, 0, 0, 0)
        self.window.ui.nodes["dialog.url.loader.label"] = HelpLabel(trans("tool.indexer.tab.web.loader"))
        self.window.ui.add_hook("update.dialog.url.web.loader", self.hook_loader_change)

        self.window.ui.nodes["dialog.url.options.label"] = HelpLabel(trans("tool.indexer.tab.web.source"))

        config_label = HelpLabel(trans("tool.indexer.tab.web.cfg"))
        config_label.setWordWrap(False)

        config_label_layout = QHBoxLayout()
        config_label_layout.addWidget(IconLabel(QIcon(":/icons/settings_filled.svg")))
        config_label_layout.addWidget(config_label)
        config_label_layout.setAlignment(Qt.AlignLeft)

        self.window.ui.nodes["dialog.url.config.label"] = QWidget()
        self.window.ui.nodes["dialog.url.config.label"].setLayout(config_label_layout)

        self.window.ui.nodes["dialog.url.config.help"] = UrlLabel(
            trans("tool.indexer.tab.web.help"),
            "https://pygpt.readthedocs.io/en/latest/configuration.html#data-loaders")

        params_layout = QVBoxLayout()
        params_layout.setContentsMargins(0, 0, 0, 0)

        self.params_scroll = QScrollArea()
        self.params_scroll.setWidgetResizable(True)
        self.window.ui.nodes["dialog.url.loader.option"] = {}
        self.window.ui.nodes["dialog.url.loader.option_group"] = {}
        self.window.ui.nodes["dialog.url.loader.config"] = {}
        self.window.ui.nodes["dialog.url.loader.config_group"] = {}

        # params
        params_layout.addWidget(self.window.ui.nodes["dialog.url.options.label"])
        inputs, groups = self.window.core.idx.ui.loaders.setup_loader_options()

        for loader in inputs:
            for k in inputs[loader]:
                self.window.ui.nodes["dialog.url.loader.option." + loader + "." + k] = inputs[loader][k]
        for loader in groups:
            self.window.ui.nodes["dialog.url.loader.option_group"][loader] = groups[loader]
            params_layout.addWidget(self.window.ui.nodes["dialog.url.loader.option_group"][loader])
            self.window.ui.nodes["dialog.url.loader.option_group"][loader].hide()  # hide on start

        # separator
        params_layout.addWidget(QHLine())

        # config
        params_layout.addWidget(self.window.ui.nodes["dialog.url.config.label"])
        inputs, groups = self.window.core.idx.ui.loaders.setup_loader_config()

        for loader in inputs:
            for k in inputs[loader]:
                self.window.ui.nodes["dialog.url.loader.config." + loader + "." + k] = inputs[loader][k]
        for loader in groups:
            self.window.ui.nodes["dialog.url.loader.config_group"][loader] = groups[loader]
            params_layout.addWidget(self.window.ui.nodes["dialog.url.loader.config_group"][loader])
            self.window.ui.nodes["dialog.url.loader.config_group"][loader].hide()  # hide on start
        params_layout.addWidget(self.window.ui.nodes["dialog.url.config.help"], alignment=Qt.AlignCenter)

        # stretch
        params_layout.addStretch(1)

        self.params_widget = QWidget()
        self.params_widget.setLayout(params_layout)
        self.params_scroll.setWidget(self.params_widget)

        # resize
        self.params_scroll.setWidgetResizable(True)
        self.params_scroll.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)

        # -----------------

        layout = QVBoxLayout()
        #layout.addWidget(self.window.ui.nodes['dialog.url.label'])
        #layout.addWidget(self.input)
        #layout.addWidget(self.window.ui.nodes['dialog.url.tip'])
        layout.addWidget(self.window.ui.nodes['dialog.url.loader.label'])
        layout.addWidget(self.window.ui.nodes['dialog.url.loader'])
        layout.addWidget(self.params_scroll)
        layout.addLayout(bottom)

        # defaults
        self.window.ui.nodes["dialog.url.loader"].set_value("webpage")

        self.setLayout(layout)

        self.initialized = True

    def hook_loader_change(self, key, value, caller, *args, **kwargs):
        """
        Hook: on loader change

        :param key: Option key
        :param value: Option value
        :param caller: Caller
        :param args: Args
        :param kwargs: Kwargs
        """
        # hide/show options
        for loader in self.window.ui.nodes["dialog.url.loader.option_group"]:
            self.window.ui.nodes["dialog.url.loader.option_group"][loader].hide()
        if value in self.window.ui.nodes["dialog.url.loader.option_group"]:
            self.window.ui.nodes["dialog.url.loader.option_group"][value].show()

        # show/hide label if options are available
        self.window.ui.nodes["dialog.url.options.label"].hide()
        if value in self.window.ui.nodes["dialog.url.loader.option_group"]:
            self.window.ui.nodes["dialog.url.options.label"].show()

        # hide/show config
        for loader in self.window.ui.nodes["dialog.url.loader.config_group"]:
            self.window.ui.nodes["dialog.url.loader.config_group"][loader].hide()
        if value in self.window.ui.nodes["dialog.url.loader.config_group"]:
            self.window.ui.nodes["dialog.url.loader.config_group"][value].show()

        # show/hide label if config are available
        self.window.ui.nodes["dialog.url.config.label"].hide()
        self.window.ui.nodes["dialog.url.config.help"].hide()
        if value in self.window.ui.nodes["dialog.url.loader.config_group"]:
            self.window.ui.nodes["dialog.url.config.label"].show()
            self.window.ui.nodes["dialog.url.config.help"].show()

        self.params_widget.adjustSize()
        self.params_scroll.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtWidgets import QApplication

from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.ctx.reply import ReplyContext


class Stack:
    def __init__(self, window=None):
        """
        Reply stack controller

        :param window: Window instance
        """
        self.window = window
        self.current = None
        self.locked = False
        self.processed = False

    def add(self, context: ReplyContext):
        """
        Add reply to stack

        :param context: ReplyContext
        """
        self.current = context
        self.unlock()
        self.processed = False

    def has(self):
        """
        Check if reply stack has any item

        :return: True if has
        """
        return self.current is not None

    def clear(self):
        """
        Clear reply stack
        """
        self.current = None
        self.unlock()

    def execute(self, context: ReplyContext):
        """
        Execute reply context

        :param context: ReplyContext
        """
        # expert call
        if context.type == ReplyContext.EXPERT_CALL:
            self.window.core.experts.call(
                context.ctx,  # master ctx
                context.parent_id,  # expert id
                context.input,  # query
            )
        elif context.type == ReplyContext.CMD_EXECUTE:
            self.window.controller.plugins.apply_cmds(
                context.ctx,  # current ctx
                context.cmds,  # commands
            )
        elif context.type == ReplyContext.CMD_EXECUTE_INLINE:
            self.window.controller.plugins.apply_cmds_inline(
                context.ctx,  # current ctx
                context.cmds,  # commands
            )
        elif context.type == ReplyContext.AGENT_CONTINUE:
            bridge_context = BridgeContext()
            bridge_context.ctx = context.ctx
            bridge_context.prompt = context.input
            extra = {
                "force": True,
                "internal": True,
            }
            event = KernelEvent(KernelEvent.INPUT_SYSTEM, {
                'context': bridge_context,
                'extra': extra,
            })
            self.window.dispatch(event)

    def is_locked(self) -> bool:
        """
        Check if reply stack is locked

        :return: True if locked
        """
        return self.locked

    def lock(self):
        """Lock reply stack"""
        self.locked = True

    def unlock(self):
        """Unlock reply stack"""
        self.locked = False

    def handle(self):
        """Handle reply stack"""
        if self.window.controller.kernel.stopped():
            self.clear()
            return

        if self.waiting():
            self.lock()
            QApplication.processEvents()
            self.execute(self.current)
            self.processed = True

    def waiting(self) -> bool:
        """
        Check if reply stack is waiting

        :return: True if waiting
        """
        return self.has() and not self.is_locked()

    def log(self, data: Any):
        """
        Log data to debug

        :param data: Data to log
        """
        self.window.core.debug.info(data)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.09 23:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QTabWidget, QMenu
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class InputTabs(QTabWidget):
    def __init__(self, window=None):
        super(InputTabs, self).__init__(window)
        self.window = window

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: QMouseEvent
        """
        if event.button() == Qt.RightButton:
            if self.tabBar().tabAt(event.pos()) == 1:  # attachments tab
                self.show_context_menu(event.globalPos())

        super(InputTabs, self).mousePressEvent(event)

    def show_context_menu(self, global_pos):
        """
        Show context menu for attachments tab

        :param global_pos: QPoint
        """
        context_menu = QMenu(self)
        actions = {}
        actions['clear'] = QAction(QIcon(":/icons/delete.svg"), trans('attachments.btn.clear'), self)
        actions['clear'].triggered.connect(
            lambda: self.window.controller.attachment.clear())
        context_menu.addAction(actions['clear'])
        context_menu.exec(global_pos)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 22:00:00                  #
# ================================================== #

from typing import Optional, Any, Dict, List

from PySide6.QtCore import QObject, QRunnable
from typing_extensions import deprecated

from .plugin import BasePlugin
from .signals import BaseSignals


class BaseWorker(QObject, QRunnable):
    def __init__(self, plugin: Optional[BasePlugin] = None, *args, **kwargs):
        QObject.__init__(self)
        QRunnable.__init__(self)
        self.plugin = plugin
        self.window = None
        self.signals = BaseSignals()
        self.args = args
        self.kwargs = kwargs
        self.cmds = None
        self.ctx = None

    def debug(self, msg: str):
        """
        Emit debug signal

        :param msg: debug message
        """
        if self.signals is not None and hasattr(self.signals, "debug"):
            self.signals.debug.emit(msg)

    def destroyed(self):
        """Emit destroyed signal"""
        if self.signals is not None and hasattr(self.signals, "destroyed"):
            self.signals.destroyed.emit()

    def error(self, err: Any):
        """
        Emit error signal

        :param err: error message
        """
        if self.signals is not None and hasattr(self.signals, "error"):
            self.signals.error.emit(err)

    def log(self, msg: str):
        """
        Emit log signal

        :param msg: log message
        """
        if self.is_threaded():
            return
        if self.signals is not None and hasattr(self.signals, "log"):
            self.signals.log.emit(msg)

    @deprecated("From 2.1.29: BaseWorker.response() is deprecated, use BaseWorker.reply() instead")
    def response(
            self,
            response: Dict[str, Any],
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Emit finished signal (deprecated)

        :param response: response (dict)
        :param extra_data: extra data
        """
        # show warning
        self.debug("BaseWorker.response is deprecated from 2.1.29, use BaseWorker.reply instead")
        self.reply(response, extra_data)

    def reply(
            self,
            response: Dict[str, Any],
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Emit finished signal (on reply from command output)

        :param response: response (dict)
        :param extra_data: extra data
        """
        # if tool call from agent_llama mode, then send direct reply to plugin -> dispatcher -> reply
        if self.ctx is not None and self.ctx.agent_call and self.plugin is not None:
            self.plugin.handle_finished(response, self.ctx, extra_data)
            return

        if self.signals is not None and hasattr(self.signals, "finished"):
            self.signals.finished.emit(response, self.ctx, extra_data)

    def reply_more(
            self,
            responses: List[Dict[str, Any]],
            extra_data: Optional[Dict[str, Any]] = None
    ):
        """
        Emit finished_more signal (on reply from command output, multiple responses)

        :param responses: list of responses dicts  TODO: add ResponseContext
        :param extra_data: extra data
        """
        # if tool call from agent_llama mode, then send direct reply to plugin -> dispatcher -> reply
        if self.ctx.agent_call and self.plugin is not None:
            self.plugin.handle_finished_more(responses, self.ctx, extra_data)
            return

        if self.signals is not None and hasattr(self.signals, "finished_more"):
            self.signals.finished_more.emit(responses, self.ctx, extra_data)

    def started(self):
        """Emit started signal"""
        if self.signals is not None and hasattr(self.signals, "started"):
            self.signals.started.emit()

    def status(self, msg: str):
        """
        Emit status signal

        :param msg: status message
        """
        if self.is_threaded():
            return
        if self.signals is not None and hasattr(self.signals, "status"):
            self.signals.status.emit(msg)

    def stopped(self):
        """Emit stopped signal"""
        if self.signals is not None and hasattr(self.signals, "stopped"):
            self.signals.stopped.emit()

    def is_threaded(self) -> bool:
        """
        Check if plugin is threaded

        :return: True if threaded
        """
        if self.plugin is not None:
            return self.plugin.is_threaded()
        return False

    def from_defaults(self, parent: BasePlugin):
        """
        Initialize defaults from parent plugin

        :param parent: parent plugin
        """
        self.plugin = parent
        self.window = parent.window

        # connect base signals
        self.signals.finished.connect(parent.handle_finished)
        self.signals.finished_more.connect(parent.handle_finished_more)
        self.signals.log.connect(parent.handle_log)
        self.signals.debug.connect(parent.handle_debug)
        self.signals.status.connect(parent.handle_status)
        self.signals.error.connect(parent.handle_error)

    def from_request(
            self,
            item: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        return {"cmd": item["cmd"]}

    def make_response(
            self,
            item: Dict[str, Any],
            result: Any,
            extra: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        """
        Prepare response item

        :param item: item with parameters
        :param result: result
        :param extra: extra data
        :return: response item
        """
        request = self.from_request(item)
        response = {
            "request": request,
            "result": result,
        }
        if extra:
            response.update(extra)
        return response

    def throw_error(self, e: Exception) -> str:
        """
        Handle error

        :param e: exception
        """
        msg = "Error: {}".format(e)
        self.error(e)
        self.log(msg)
        return msg

    def has_param(
            self,
            item: Dict[str, Any],
            param: str
    ) -> bool:
        """
        Check if item has parameter

        :param item: item with parameters
        :param param: parameter name
        :return: True if item has parameter
        """
        if item is None:
            return False
        return "params" in item and param in item["params"]

    def get_param(
            self,
            item: Dict[str, Any],
            param: str,
            default: Any = None
    ) -> Any:
        """
        Get parameter value from item

        :param item: item with parameters
        :param param: parameter name
        :param default: default value
        :return: parameter value
        """
        if self.has_param(item, param):
            return item["params"][param]
        return default

    def is_stopped(self) -> bool:
        """
        Check if worker is stopped

        :return: True if stopped
        """
        if self.plugin is not None:
            return self.plugin.window.controller.kernel.stopped()
        return False

    def run_sync(self):
        """Run synchronous"""
        self.run()

    def run_async(self):
        """Run asynchronous"""
        if self.window:
            self.window.threadpool.start(self)
        else:
            self.run()
